import time
from queue import Queue

import pytest
from eth_pydantic_types import HexBytes

from ape.exceptions import ChainError, ProviderError


def test_iterate_blocks(chain_that_mined_5):
    blocks = [b for b in chain_that_mined_5.blocks]
    assert len(blocks) >= 5  # Because mined 5 blocks so is at least 5

    iterator = blocks[0].number
    for block in blocks:
        assert block.number == iterator
        iterator += 1


def test_blocks_range(chain_that_mined_5):
    # The number of the block before mining the 5
    start_block = len(chain_that_mined_5.blocks) - 5
    num_to_get = 3  # Expecting blocks [s, s+1, s+2]
    blocks = [b for b in chain_that_mined_5.blocks.range(start_block, start_block + num_to_get)]
    assert len(blocks) == num_to_get

    expected_number = start_block
    prev_block_hash = (
        HexBytes("0x0000000000000000000000000000000000000000000000000000000000000000")
        if start_block == 0
        else chain_that_mined_5.blocks[start_block - 1].hash
    )
    for block in blocks:
        assert block.number == expected_number
        expected_number += 1
        assert block.parent_hash == prev_block_hash
        prev_block_hash = block.hash


def test_blocks_range_too_high_stop(chain_that_mined_5):
    num_blocks = len(chain_that_mined_5.blocks)
    num_blocks_add_1 = num_blocks + 1
    expected = (
        rf"'stop={num_blocks_add_1}' cannot be greater than the chain length \({num_blocks}\)\. "
        rf"Use 'poll_blocks\(\)' to wait for future blocks\."
    )
    with pytest.raises(ChainError, match=expected):
        # Have to run through generator to trigger code in definition.
        list(chain_that_mined_5.blocks.range(num_blocks_add_1))


def test_block_range_with_step(chain_that_mined_5):
    blocks = [b for b in chain_that_mined_5.blocks.range(3, step=2)]
    assert len(blocks) == 2
    assert blocks[0].number == 0
    assert blocks[1].number == 2


def test_block_range_negative_start(chain_that_mined_5):
    with pytest.raises(ValueError) as err:
        _ = [b for b in chain_that_mined_5.blocks.range(-1, 3, step=2)]

    assert "Input should be greater than or equal to 0" in str(err.value)


def test_block_range_out_of_order(chain_that_mined_5):
    with pytest.raises(ValueError) as err:
        _ = [b for b in chain_that_mined_5.blocks.range(3, 1, step=2)]

    assert "stop_block: '0' cannot be less than start_block: '3'." in str(err.value)


def test_block_timestamp(chain):
    chain.mine()
    assert chain.blocks.head.timestamp == chain.blocks.head.datetime.timestamp()


def test_poll_blocks_stop_block_not_in_future(chain_that_mined_5):
    bad_stop_block = chain_that_mined_5.blocks.height

    with pytest.raises(ValueError, match="'stop' argument must be in the future."):
        _ = [x for x in chain_that_mined_5.blocks.poll_blocks(stop_block=bad_stop_block)]


def test_poll_blocks(chain_that_mined_5, eth_tester_provider, owner, PollDaemon):
    blocks: Queue = Queue(maxsize=3)
    poller = chain_that_mined_5.blocks.poll_blocks()

    with PollDaemon("blocks", poller, blocks.put, blocks.full):
        # Sleep first to ensure listening before mining.
        time.sleep(1)
        eth_tester_provider.mine(3)

    assert blocks.full()
    first = blocks.get().number
    second = blocks.get().number
    third = blocks.get().number
    assert first == second - 1
    assert second == third - 1


def test_poll_blocks_reorg(chain_that_mined_5, eth_tester_provider, owner, PollDaemon, ape_caplog):
    blocks: Queue = Queue(maxsize=6)
    poller = chain_that_mined_5.blocks.poll_blocks()

    with PollDaemon("blocks", poller, blocks.put, blocks.full):
        # Sleep first to ensure listening before mining.
        time.sleep(1)

        snapshot = chain_that_mined_5.snapshot()
        chain_that_mined_5.mine(2)

        # Wait to allow blocks before re-org to get yielded
        time.sleep(5)

        # Simulate re-org by reverting to the snapshot
        chain_that_mined_5.restore(snapshot)

        # Allow it time to trigger realizing there was a re-org
        time.sleep(1)
        chain_that_mined_5.mine(2)
        time.sleep(1)

        chain_that_mined_5.mine(3)

    assert blocks.full()

    # Show that re-org was detected
    expected_error = (
        "Chain has reorganized since returning the last block. "
        "Try adjusting the required network confirmations."
    )
    assert ape_caplog.records, "Didn't detect re-org"
    ape_caplog.assert_last_log(expected_error)

    # Show that there are duplicate blocks
    block_numbers: list[int] = [blocks.get().number for _ in range(6)]
    assert len(set(block_numbers)) < len(block_numbers)


def test_poll_blocks_timeout(
    vyper_contract_instance, chain_that_mined_5, eth_tester_provider, owner, PollDaemon
):
    poller = chain_that_mined_5.blocks.poll_blocks(new_block_timeout=1)

    with pytest.raises(ProviderError, match=r"Timed out waiting for next block."):
        with PollDaemon("blocks", poller, lambda x: None, lambda: False):
            time.sleep(1.5)


def test_poll_blocks_reorg_with_timeout(
    vyper_contract_instance, chain_that_mined_5, eth_tester_provider, owner, PollDaemon, ape_caplog
):
    blocks: Queue = Queue(maxsize=6)
    poller = chain_that_mined_5.blocks.poll_blocks(new_block_timeout=1)

    with pytest.raises(ProviderError, match=r"Timed out waiting for next block."):
        with PollDaemon("blocks", poller, blocks.put, blocks.full):
            # Sleep first to ensure listening before mining.
            time.sleep(1)

            snapshot = chain_that_mined_5.snapshot()
            chain_that_mined_5.mine(2)

            # Wait to allow blocks before re-org to get yielded
            time.sleep(5)

            # Simulate re-org by reverting to the snapshot
            chain_that_mined_5.restore(snapshot)

            # Allow it time to trigger realizing there was a re-org
            time.sleep(1)
            chain_that_mined_5.mine(2)
            time.sleep(1)

            chain_that_mined_5.mine(3)


def test_poll_blocks_future(chain_that_mined_5, eth_tester_provider, owner, PollDaemon):
    blocks: Queue = Queue(maxsize=3)
    poller = chain_that_mined_5.blocks.poll_blocks(
        start_block=chain_that_mined_5.blocks.head.number + 1
    )

    with PollDaemon("blocks", poller, blocks.put, blocks.full):
        # Sleep first to ensure listening before mining.
        time.sleep(1)
        eth_tester_provider.mine(3)

    assert blocks.full()
    first = blocks.get().number
    second = blocks.get().number
    third = blocks.get().number
    assert first == second - 1
    assert second == third - 1



import json
import sys
from abc import abstractmethod
from collections import defaultdict, deque
from collections.abc import Iterable, Iterator, Sequence
from enum import Enum
from functools import cached_property
from typing import IO, TYPE_CHECKING, Any, Optional, Union

from eth_pydantic_types import HexStr
from eth_utils import is_0x_prefixed, to_hex
from evm_trace import (
    CallTreeNode,
    CallType,
    ParityTraceList,
    TraceFrame,
    create_trace_frames,
    get_calltree_from_geth_call_trace,
    get_calltree_from_geth_trace,
    get_calltree_from_parity_trace,
)
from evm_trace.gas import merge_reports
from hexbytes import HexBytes
from pydantic import field_validator
from rich.tree import Tree

from ape.api.trace import TraceAPI
from ape.api.transactions import TransactionAPI
from ape.exceptions import ContractLogicError, ProviderError, TransactionNotFoundError
from ape.logging import get_rich_console, logger
from ape.utils.misc import ZERO_ADDRESS, is_evm_precompile, is_zero_hex, log_instead_of_fail
from ape.utils.trace import TraceStyles, _exclude_gas
from ape_ethereum._print import extract_debug_logs

if TYPE_CHECKING:
    from ethpm_types import ContractType, MethodABI

    from ape.api.networks import EcosystemAPI
    from ape.types.address import AddressType
    from ape.types.trace import ContractFunctionPath, GasReport


_INDENT = 2
_WRAP_THRESHOLD = 50
_REVERT_PREFIX = "0x08c379a00000000000000000000000000000000000000000000000000000000000000020"


class TraceApproach(Enum):
    """RPC trace_transaction."""

    BASIC = 0
    """No tracing support; think of EthTester."""

    PARITY = 1
    """RPC 'trace_transaction'."""

    GETH_CALL_TRACER = 2
    """RPC debug_traceTransaction using tracer='callTracer'."""

    GETH_STRUCT_LOG_PARSE = 3
    """
    RPC debug_traceTransaction using struct-log tracer
    and sophisticated parsing from the evm-trace library.
    NOT RECOMMENDED.
    """

    @classmethod
    def from_key(cls, key: str) -> "TraceApproach":
        return cls(cls._validate(key))

    @classmethod
    def _validate(cls, key: Any) -> "TraceApproach":
        if isinstance(key, TraceApproach):
            return key
        elif isinstance(key, int) or (isinstance(key, str) and key.isnumeric()):
            return cls(int(key))

        # Check if given a name.
        key = key.replace("-", "_").upper()

        # Allow shorter, nicer values for the geth-struct-log approach.
        if key in ("GETH", "GETH_STRUCT_LOG", "GETH_STRUCT_LOGS"):
            key = "GETH_STRUCT_LOG_PARSE"

        for member in cls:
            if member.name == key:
                return member

        raise ValueError(f"No enum named '{key}'.")


class Trace(TraceAPI):
    """
    Set to ``True`` to use an ERC-20's SYMBOL as the contract's identifier.
    Is ``True`` when showing pretty traces without gas tables. When gas is
    involved, Ape must use the ``.name`` as the identifier for all contracts.
    """

    call_trace_approach: Optional[TraceApproach] = None
    """When None, attempts to deduce."""

    _enriched_calltree: Optional[dict] = None

    def __repr__(self) -> str:
        try:
            return f"{self}"
        except Exception as err:
            # Don't let __repr__ fail.
            logger.debug(f"Problem transaction trace: {err}")
            return "<Trace>"

    def __str__(self) -> str:
        return _call_to_str(self.enriched_calltree)

    @log_instead_of_fail()
    def _repr_pretty_(self, *args, **kwargs):
        self.show()

    @property
    @abstractmethod
    def raw_trace_frames(self) -> Iterator[dict]:
        """
        The raw trace frames.
        """

    @property
    @abstractmethod
    def transaction(self) -> dict:
        """
        The transaction data (obtained differently on
        calls versus transactions).
        """

    @abstractmethod
    def get_calltree(self) -> CallTreeNode:
        """
        Get an un-enriched call-tree node.
        """

    @cached_property
    def debug_logs(self) -> Iterable[tuple[Any]]:
        """
        Calls from ``console.log()`` and ``print()`` from a transactions call tree.
        """
        return list(extract_debug_logs(self.get_calltree()))

    @property
    def enriched_calltree(self) -> dict:
        """
        The fully enriched calltree node.
        """
        if self._enriched_calltree is not None:
            return self._enriched_calltree

        # Side-effect: sets `_enriched_calltree` if using Ethereum node provider.
        self.provider.network.ecosystem.enrich_trace(self)

        if self._enriched_calltree is None:
            # If still None (shouldn't be), set to avoid repeated attempts.
            self._enriched_calltree = {}

        # Add top-level data if missing.
        if not self._enriched_calltree.get("gas_cost"):
            # Happens on calltrees built from structLogs.
            if gas_used := self.transaction.get("gas_used"):
                if "data" in self.transaction:
                    # Subtract base gas costs.
                    # (21_000 + 4 gas per 0-byte and 16 gas per non-zero byte).
                    data_gas = sum(
                        [4 if x == 0 else 16 for x in HexBytes(self.transaction["data"])]
                    )
                    self._enriched_calltree["gas_cost"] = gas_used - 21_000 - data_gas

        return self._enriched_calltree

    @property
    def frames(self) -> Iterator[TraceFrame]:
        yield from create_trace_frames(iter(self.raw_trace_frames))

    @property
    def addresses(self) -> Iterator["AddressType"]:
        yield from self.get_addresses_used()

    @cached_property
    def root_contract_type(self) -> Optional["ContractType"]:
        if address := self.transaction.get("to"):
            try:
                return self.chain_manager.contracts.get(address)
            except Exception:
                return None

        return None

    @cached_property
    def root_method_abi(self) -> Optional["MethodABI"]:
        method_id = self.transaction.get("data", b"")[:10]
        if ct := self.root_contract_type:
            try:
                return ct.methods[method_id]
            except Exception:
                return None

        return None

    @property
    def _ecosystem(self) -> "EcosystemAPI":
        if provider := self.network_manager.active_provider:
            return provider.network.ecosystem

        # Default to Ethereum (since we are in that plugin!)
        return self.network_manager.ethereum

    def get_addresses_used(self, reverse: bool = False):
        frames: Iterable
        if reverse:
            frames = list(self.frames)
            frames = frames[::-1] if reverse else frames
        else:
            # Don't need to run whole list.
            frames = self.frames

        for frame in frames:
            if not (addr := frame.address):
                continue

            yield self._ecosystem.decode_address(addr)

    @cached_property
    def return_value(self) -> Any:
        if self._enriched_calltree:
            # Only check enrichment output if was already enriched!
            # Don't enrich ONLY for return value, as that is very bad performance
            # for realistic contract interactions.
            return self._return_value_from_enriched_calltree

        # Barely enrich a calltree for performance reasons
        # (likely not a need to enrich the whole thing).
        calltree = self.get_calltree()
        return self._get_return_value_from_calltree(calltree)

    @cached_property
    def _return_value_from_enriched_calltree(self) -> Any:
        calltree = self.enriched_calltree

        # Check if was cached from enrichment.
        if "return_value" in self.__dict__:
            return self.__dict__["return_value"]

        return self._get_return_value_from_calltree(calltree)

    def _get_return_value_from_calltree(
        self, calltree: Union[dict, CallTreeNode]
    ) -> tuple[Optional[Any], ...]:
        num_outputs = 1
        if raw_return_data := (
            calltree.get("returndata") if isinstance(calltree, dict) else calltree.returndata
        ):
            if abi := self._get_abi(calltree):
                # Ensure we return a tuple with the correct length, even if fails.
                num_outputs = len(abi.outputs)
                try:
                    return self._ecosystem.decode_returndata(abi, HexBytes(raw_return_data))
                except Exception as err:
                    logger.debug(
                        f"Failed decoding raw returndata: {to_hex(raw_return_data)}. Error: {err}"
                    )
                    return tuple([None for _ in range(num_outputs)])

        return tuple([None for _ in range(num_outputs)])

    @cached_property
    def revert_message(self) -> Optional[str]:
        call = self.enriched_calltree
        if not call.get("failed", False):
            return None

        def try_get_revert_msg(c) -> Optional[str]:
            if msg := c.get("revert_message"):
                return msg

            for sub_c in c.get("calls", []):
                if msg := try_get_revert_msg(sub_c):
                    return msg

            return None

        if message := try_get_revert_msg(call):
            return message

        # Enrichment call-tree not available. Attempt looking in trace-frames.
        if revert_str := self._revert_str_from_trace_frames:
            return to_hex(revert_str)

        return None

    @cached_property
    def _last_frame(self) -> Optional[dict]:
        try:
            frame = deque(self.raw_trace_frames, maxlen=1)
        except Exception as err:
            logger.error(f"Failed getting traceback: {err}")
            return None

        return frame[0] if frame else None

    @cached_property
    def _revert_str_from_trace_frames(self) -> Optional[HexBytes]:
        if frame := self._last_frame:
            memory = frame.get("memory", [])
            if ret := "".join([x[2:] for x in memory[4:]]):
                return HexBytes(ret)

        return None

    @cached_property
    def _return_data_from_trace_frames(self) -> Optional[HexBytes]:
        if frame := self._last_frame:
            memory = frame["memory"]
            start_pos = int(frame["stack"][2], 16) // 32
            return HexBytes("".join(memory[start_pos:]))

        return None

    """ API Methods """

    def show(self, verbose: bool = False, file: IO[str] = sys.stdout):
        call = self.enriched_calltree
        approaches_handling_events = (TraceApproach.GETH_STRUCT_LOG_PARSE,)

        failed = call.get("failed", False)
        revert_message = None
        if failed:
            revert_message = self.revert_message
            revert_message = (
                f'reverted with message: "{revert_message}"'
                if revert_message
                else "reverted without message"
            )

        root = self._get_tree(verbose=verbose)
        console = get_rich_console(file=file)
        if txn_hash := getattr(self, "transaction_hash", None):
            # Only works on TransactionTrace (not CallTrace).
            console.print(f"Call trace for [bold blue]'{txn_hash}'[/]")

        if revert_message:
            console.print(f"[bold red]{revert_message}[/]")

        if sender := self.transaction.get("from"):
            console.print(f"tx.origin=[{TraceStyles.CONTRACTS}]{sender}[/]")

        if self.call_trace_approach not in approaches_handling_events and hasattr(
            self._ecosystem, "_enrich_trace_events"
        ):
            # We must manually attach the contract logs.
            # NOTE: With these approaches, we don't know where they appear
            #   in the call-tree so we have to put them at the top.
            if logs := self.transaction.get("logs", []):
                enriched_events = self._ecosystem._enrich_trace_events(logs)
                event_trees = _events_to_trees(enriched_events)
                if event_trees:
                    console.print()
                    self.chain_manager._reports.show_events(event_trees, console=console)
                    console.print()

        # else: the events are already included in the right spots in the call tree.

        console.print(root)

    def get_gas_report(
        self, exclude: Optional[Sequence["ContractFunctionPath"]] = None
    ) -> "GasReport":
        call = self.enriched_calltree
        return self._get_gas_report_from_call(call, exclude=exclude)

    def _get_gas_report_from_call(
        self, call: dict, exclude: Optional[Sequence["ContractFunctionPath"]] = None
    ) -> "GasReport":
        tx = self.transaction

        # Enrich transfers.
        contract_id = call.get("contract_id", "")
        is_transfer = contract_id.startswith("__") and contract_id.endswith("transfer__")
        if is_transfer and tx.get("to") is not None and tx["to"] in self.account_manager:
            receiver_id = self.account_manager[tx["to"]].alias or tx["to"]
            call["method_id"] = f"to:{receiver_id}"

        elif is_transfer and (receiver := tx.get("to")):
            call["method_id"] = f"to:{receiver}"

        exclusions = exclude or []
        calls = call.get("calls", [])
        sub_reports = (self._get_gas_report_from_call(c, exclude=exclusions) for c in calls)

        if (
            not call.get("contract_id")
            or not call.get("method_id")
            or _exclude_gas(exclusions, call.get("contract_id", ""), call.get("method_id", ""))
        ):
            return merge_reports(*sub_reports)

        elif not is_zero_hex(call["method_id"]) and not is_evm_precompile(call["method_id"]):
            report: GasReport = {
                call["contract_id"]: {
                    call["method_id"]: (
                        [int(call["gas_cost"])] if call.get("gas_cost") is not None else []
                    )
                }
            }
            return merge_reports(*sub_reports, report)

        return merge_reports(*sub_reports)

    def show_gas_report(self, verbose: bool = False, file: IO[str] = sys.stdout):
        gas_report = self.get_gas_report()
        self.chain_manager._reports.show_gas(gas_report, file=file)

    def get_raw_frames(self) -> Iterator[dict]:
        yield from self.raw_trace_frames

    def get_raw_calltree(self) -> dict:
        return self.get_calltree().model_dump(mode="json", by_alias=True)

    """ Shared helpers """

    def _get_tx_calltree_kwargs(self) -> dict:
        if (receiver := self.transaction.get("to")) and receiver != ZERO_ADDRESS:
            call_type = CallType.CALL
        else:
            call_type = CallType.CREATE
            receiver = self.transaction.get("contract_address")

        return {
            "address": receiver,
            "call_type": call_type,
            "calldata": self.transaction.get("data", b""),
            "gas_cost": self.transaction.get("gasCost"),
            "failed": False,
            "value": self.transaction.get("value", 0),
        }

    def _debug_trace_transaction_struct_logs_to_call(self) -> CallTreeNode:
        init_kwargs = self._get_tx_calltree_kwargs()
        return get_calltree_from_geth_trace(self.frames, **init_kwargs)

    def _get_tree(self, verbose: bool = False) -> Tree:
        return parse_rich_tree(self.enriched_calltree, verbose=verbose)

    def _get_abi(self, call: Union[dict, CallTreeNode]) -> Optional["MethodABI"]:
        if not (addr := call.get("address") if isinstance(call, dict) else call.address):
            return self.root_method_abi
        if not (calldata := call.get("calldata") if isinstance(call, dict) else call.calldata):
            return self.root_method_abi
        if not (contract_type := self.chain_manager.contracts.get(addr)):
            return self.root_method_abi
        if calldata[:10] not in contract_type.methods:
            return self.root_method_abi

        return contract_type.methods[calldata[:10]]


class TransactionTrace(Trace):
    transaction_hash: HexStr
    debug_trace_transaction_parameters: dict = {"enableMemory": True}
    _frames: list[dict] = []

    @property
    def raw_trace_frames(self) -> Iterator[dict]:
        """
        The raw trace ``"structLogs"`` from ``debug_traceTransaction``
        for deeper investigation.
        """
        if self._frames:
            yield from self._frames

        else:
            for frame in self._stream_struct_logs():
                self._frames.append(frame)
                yield frame

    @cached_property
    def transaction(self) -> dict:
        receipt = self.chain_manager.get_receipt(self.transaction_hash)
        data = receipt.transaction.model_dump(mode="json", by_alias=True)
        return {**data, **receipt.model_dump(by_alias=True)}

    def _stream_struct_logs(self) -> Iterator[dict]:
        parameters = self.debug_trace_transaction_parameters
        yield from self.provider.stream_request(
            "debug_traceTransaction",
            [self.transaction_hash, parameters],
            iter_path="result.structLogs.item",
        )

    def get_calltree(self) -> CallTreeNode:
        if self.call_trace_approach is TraceApproach.BASIC:
            return self._get_basic_calltree()

        elif self.call_trace_approach is TraceApproach.PARITY:
            return self._trace_transaction()

        elif self.call_trace_approach is TraceApproach.GETH_CALL_TRACER:
            return self._debug_trace_transaction_call_tracer()

        elif self.call_trace_approach is TraceApproach.GETH_STRUCT_LOG_PARSE:
            return self._debug_trace_transaction_struct_logs_to_call()

        elif "erigon" in getattr(self.provider, "client_version", "").lower():
            # Based on the client version, we know parity works.
            call = self._trace_transaction()
            self._set_approach(TraceApproach.PARITY)
            return call

        return self._discover_calltrace_approach()

    def _discover_calltrace_approach(self) -> CallTreeNode:
        # NOTE: This method is only called once, if at all.
        #   After discovery, short-circuits to the correct approach.
        #   It tries to create an evm_trace.CallTreeNode using
        #   all the approaches in order from fastest to slowest.

        TA = TraceApproach
        approaches = {
            TA.PARITY: self._trace_transaction,
            TA.GETH_CALL_TRACER: self._debug_trace_transaction_call_tracer,
            TA.GETH_STRUCT_LOG_PARSE: self._debug_trace_transaction_struct_logs_to_call,
            TA.BASIC: self._get_basic_calltree,
        }

        reason_map = {}
        for approach, fn in approaches.items():
            try:
                call = fn()
            except Exception as err:
                reason_map[approach.name] = f"{err}"
                continue

            self._set_approach(approach)
            return call

        # Not sure this happens, as the basic-approach should always work.
        reason_str = ", ".join(f"{k}={v}" for k, v in reason_map.items())
        raise ProviderError(f"Unable to create CallTreeNode. Reason(s): {reason_str}")

    def _debug_trace_transaction(self, parameters: Optional[dict] = None) -> dict:
        parameters = parameters or self.debug_trace_transaction_parameters
        return self.provider.make_request(
            "debug_traceTransaction", [self.transaction_hash, parameters]
        )

    def _debug_trace_transaction_call_tracer(self) -> CallTreeNode:
        parameters = {**self.debug_trace_transaction_parameters, "tracer": "callTracer"}
        data = self._debug_trace_transaction(parameters)
        return get_calltree_from_geth_call_trace(data)

    def _trace_transaction(self) -> CallTreeNode:
        try:
            data = self.provider.make_request("trace_transaction", [self.transaction_hash])
        except ProviderError as err:
            if "transaction not found" in str(err).lower():
                raise TransactionNotFoundError(transaction_hash=self.transaction_hash) from err

            raise  # The ProviderError as-is

        parity_objects = ParityTraceList.model_validate(data)
        return get_calltree_from_parity_trace(parity_objects)

    def _get_basic_calltree(self) -> CallTreeNode:
        init_kwargs = self._get_tx_calltree_kwargs()
        receipt = self.chain_manager.get_receipt(self.transaction_hash)
        init_kwargs["gas_cost"] = receipt.gas_used

        # Figure out the 'returndata' using 'eth_call' RPC.
        tx = receipt.transaction.model_copy(update={"nonce": None})
        try:
            return_value = self.provider.send_call(tx, block_id=receipt.block_number)
        except ContractLogicError:
            # Unable to get the return value because even as a call, it fails.
            pass
        else:
            init_kwargs["returndata"] = return_value

        return CallTreeNode(**init_kwargs)

    def _set_approach(self, approach: TraceApproach):
        self.call_trace_approach = approach
        if hasattr(self.provider, "_call_trace_approach"):
            self.provider._call_trace_approach = approach


class CallTrace(Trace):
    tx: dict
    """
    Transaction data. Is a dictionary to allow traces to easily
    be created near sending the request.
    """

    arguments: list[Any] = []
    """
    Remaining eth-call arguments, minus the transaction.
    """

    call_trace_approach: TraceApproach = TraceApproach.GETH_STRUCT_LOG_PARSE
    """debug_traceCall must use the struct-log tracer."""

    supports_debug_trace_call: Optional[bool] = None

    @field_validator("tx", mode="before")
    @classmethod
    def _tx_to_dict(cls, value):
        if isinstance(value, TransactionAPI):
            return value.model_dump(by_alias=True)

        return value

    @property
    def raw_trace_frames(self) -> Iterator[dict]:
        yield from self._traced_call.get("structLogs", [])

    @property
    def return_value(self) -> Any:
        return self._traced_call.get("returnValue", "")

    @cached_property
    def _traced_call(self) -> dict:
        if self.supports_debug_trace_call is True:
            return self._debug_trace_call()
        elif self.supports_debug_trace_call is False:
            return {}

        try:
            result = self._debug_trace_call()
        except Exception:
            self._set_supports_trace_call(False)
            return {}

        self._set_supports_trace_call(True)
        return result

    @property
    def transaction(self) -> dict:
        return self.tx

    def get_calltree(self) -> CallTreeNode:
        calltree = self._debug_trace_transaction_struct_logs_to_call()
        calltree.gas_cost = self._traced_call.get("gas", calltree.gas_cost)
        calltree.failed = self._traced_call.get("failed", calltree.failed)
        return calltree

    def _set_supports_trace_call(self, value: bool):
        self.supports_debug_trace_call = value
        if hasattr(self.provider, "_supports_debug_trace_call"):
            self.provider._supports_debug_trace_call = True

    def _debug_trace_call(self):
        arguments = [self.transaction, *self.arguments]

        # Block ID is required, at least for regular geth nodes.
        if len(arguments) == 1:
            arguments.append("latest")

        return self.provider.make_request("debug_traceCall", arguments)


def parse_rich_tree(call: dict, verbose: bool = False) -> Tree:
    tree = _create_tree(call, verbose=verbose)
    for event in call.get("events", []):
        if "calldata" not in event and "name" not in event:
            # Not sure; or not worth showing.
            logger.debug(f"Unknown event data: '{event}'.")
            continue

        event_tree = _create_event_tree(event)
        tree.add(event_tree)

    for sub_call in call.get("calls", []):
        sub_tree = parse_rich_tree(sub_call, verbose=verbose)
        tree.add(sub_tree)

    return tree


def _events_to_trees(events: list[dict]) -> list[Tree]:
    event_counter = defaultdict(list)
    for evt in events:
        name = evt.get("name")
        calldata = evt.get("calldata")

        if not name or not calldata:
            # Not sure; or not worth showing.
            logger.debug(f"Unknown event data: '{evt}'.")
            continue

        tuple_key = (
            name,
            ",".join(f"{k}={v}" for k, v in calldata.items()),
        )
        event_counter[tuple_key].append(evt)

    result = []
    for evt_tup, events in event_counter.items():
        count = len(events)
        evt = events[0]
        if "name" not in evt and "calldata" not in evt:
            # Not sure; or not worth showing.
            logger.debug(f"Unknown event data: '{evt}'.")
            continue

        # NOTE: Using similar style to gas-cost on purpose.
        suffix = f"[[{TraceStyles.GAS_COST}]x{count}[/]]" if count > 1 else ""
        evt_tree = _create_event_tree(evt, suffix=suffix)
        result.append(evt_tree)

    return result


def _create_event_tree(event: dict, suffix: str = "") -> Tree:
    signature = _event_to_str(event, stylize=True, suffix=suffix)
    return Tree(signature)


def _call_to_str(call: dict, stylize: bool = False, verbose: bool = False) -> str:
    contract = str(call.get("contract_id", ""))
    is_create = "CREATE" in call.get("call_type", "")
    method = (
        "__new__"
        if is_create and call["method_id"] and is_0x_prefixed(call["method_id"])
        else str(call.get("method_id") or "")
    )
    if "(" in method:
        # Only show short name, not ID name
        # (it is the full signature when multiple methods have the same name).
        method = method.split("(")[0].strip() or method

    if stylize:
        contract = f"[{TraceStyles.CONTRACTS}]{contract}[/]"
        method = f"[{TraceStyles.METHODS}]{method}[/]"

    call_path = f"{contract}.{method}"

    if call.get("call_type") is not None and call["call_type"].upper() == "DELEGATECALL":
        delegate = "(delegate)"
        if stylize:
            delegate = f"[orange]{delegate}[/]"

        call_path = f"{delegate} {call_path}"

    arguments_str = _get_inputs_str(call.get("calldata"), stylize=stylize)
    if is_create and is_0x_prefixed(arguments_str):
        # Un-enriched CREATE calldata is a massive hex.
        arguments_str = ""

    signature = f"{call_path}{arguments_str}"
    returndata = call.get("returndata", "")
    if not is_create and returndata not in ((), [], None, {}, ""):
        if return_str := _get_outputs_str(returndata, stylize=stylize):
            signature = f"{signature} -> {return_str}"

    if call.get("value"):
        value = str(call["value"])
        if stylize:
            value = f"[{TraceStyles.VALUE}]{value}[/]"

        signature += f" {value}"

    if call.get("gas_cost"):
        gas_value = f"[{call['gas_cost']} gas]"
        if stylize:
            gas_value = f"[{TraceStyles.GAS_COST}]{gas_value}[/]"

        signature += f" {gas_value}"

    if verbose:
        verbose_items = {k: v for k, v in call.items() if type(v) in (int, str, bytes, float)}
        extra = json.dumps(verbose_items, indent=2)
        signature = f"{signature}\n{extra}"

    return signature


def _event_to_str(event: dict, stylize: bool = False, suffix: str = "") -> str:
    # NOTE: Some of the styles are matching others parts of the trace,
    #  even though the 'name' is a bit misleading.
    event_name = event.get("name", "ANONYMOUS_EVENT")
    name = f"[{TraceStyles.METHODS}]{event_name}[/]" if stylize else event_name
    arguments_str = _get_inputs_str(event.get("calldata", "0x"), stylize=stylize)
    prefix = f"[{TraceStyles.CONTRACTS}]log[/]" if stylize else "log"
    return f"{prefix} {name}{arguments_str}{suffix}"


def _create_tree(call: dict, verbose: bool = False) -> Tree:
    signature = _call_to_str(call, stylize=True, verbose=verbose)
    return Tree(signature)


def _get_inputs_str(inputs: Any, stylize: bool = False) -> str:
    color = TraceStyles.INPUTS if stylize else None
    if inputs in ["0x", None, (), [], {}]:
        return "()"

    elif isinstance(inputs, dict):
        return _dict_to_str(inputs, color=color)

    elif isinstance(inputs, bytes):
        return to_hex(inputs)

    return f"({inputs})"


def _get_outputs_str(outputs: Any, stylize: bool = False) -> Optional[str]:
    if outputs in ["0x", None, (), [], {}]:
        return None

    elif isinstance(outputs, dict):
        color = TraceStyles.OUTPUTS if stylize else None
        return _dict_to_str(outputs, color=color)

    elif isinstance(outputs, (list, tuple)):
        return (
            f"[{TraceStyles.OUTPUTS}]{_list_to_str(outputs)}[/]"
            if stylize
            else _list_to_str(outputs)
        )

    return f"[{TraceStyles.OUTPUTS}]{outputs}[/]" if stylize else str(outputs)


def _dict_to_str(dictionary: dict, color: Optional[str] = None) -> str:
    length = sum(len(str(v)) for v in [*dictionary.keys(), *dictionary.values()])
    do_wrap = length > _WRAP_THRESHOLD

    index = 0
    end_index = len(dictionary) - 1
    kv_str = "(\n" if do_wrap else "("

    for key, value in dictionary.items():
        if do_wrap:
            kv_str += _INDENT * " "

        if isinstance(value, (list, tuple)):
            value = _list_to_str(value, 1 if do_wrap else 0)

        value_str = f"[{color}]{value}[/]" if color is not None else str(value)
        kv_str += f"{key}={value_str}" if key and not key.isnumeric() else value_str
        if index < end_index:
            kv_str += ", "

        if do_wrap:
            kv_str += "\n"

        index += 1

    return f"{kv_str})"


def _list_to_str(ls: Union[list, tuple], depth: int = 0) -> str:
    if not isinstance(ls, (list, tuple)) or len(str(ls)) < _WRAP_THRESHOLD:
        return str(ls)

    elif ls and isinstance(ls[0], (list, tuple)):
        # List of lists
        sub_lists = [_list_to_str(i) for i in ls]

        # Use multi-line if exceeds threshold OR any of the sub-lists use multi-line
        extra_chars_len = (len(sub_lists) - 1) * 2
        use_multiline = len(str(sub_lists)) + extra_chars_len > _WRAP_THRESHOLD or any(
            ["\n" in ls for ls in sub_lists]
        )

        if not use_multiline:
            # Happens for lists like '[[0], [1]]' that are short.
            return f"[{', '.join(sub_lists)}]"

        value = "[\n"
        num_sub_lists = len(sub_lists)
        index = 0
        spacing = _INDENT * " " * 2
        for formatted_list in sub_lists:
            if "\n" in formatted_list:
                # Multi-line sub list. Append 1 more spacing to each line.
                indented_item = f"\n{spacing}".join(formatted_list.splitlines())
                value = f"{value}{spacing}{indented_item}"
            else:
                # Single line sub-list
                value = f"{value}{spacing}{formatted_list}"

            if index < num_sub_lists - 1:
                value = f"{value},"

            value = f"{value}\n"
            index += 1

        value = f"{value}]"
        return value

    return _list_to_multiline_str(ls, depth=depth)


def _list_to_multiline_str(value: Union[list, tuple], depth: int = 0) -> str:
    spacing = _INDENT * " "
    ls_spacing = spacing * (depth + 1)
    joined = ",\n".join([f"{ls_spacing}{v}" for v in value])
    new_val = f"[\n{joined}\n{spacing * depth}]"
    return new_val



from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

import click
from click import BadArgumentUsage

from ape.cli.choices import _ACCOUNT_TYPE_FILTER, Alias
from ape.logging import logger

if TYPE_CHECKING:
    from ape.managers.project import ProjectManager


def _alias_callback(ctx, param, value):
    from ape.utils.validators import _validate_account_alias

    return _validate_account_alias(value)


def existing_alias_argument(account_type: _ACCOUNT_TYPE_FILTER = None, **kwargs):
    """
    A ``click.argument`` for an existing account alias.

    Args:
        account_type (type[:class:`~ape.api.accounts.AccountAPI`], optional):
          If given, limits the type of account the user may choose from.
        **kwargs: click.argument overrides.
    """
    type_ = kwargs.pop("type", Alias(key=account_type))
    return click.argument("alias", type=type_, **kwargs)


def non_existing_alias_argument(**kwargs):
    """
    A ``click.argument`` for an account alias that does not yet exist in ape.

    Args:
        **kwargs: click.argument overrides.
    """

    callback = kwargs.pop("callback", _alias_callback)
    return click.argument("alias", callback=callback, **kwargs)


class _ContractPaths:
    """
    Helper callback class for handling CLI-given contract paths.
    """

    def __init__(self, value, project: Optional["ProjectManager"] = None):
        from ape.utils.basemodel import ManagerAccessMixin

        self.value = value
        self.missing_compilers: set[str] = set()  # set of .ext
        self.project = project or ManagerAccessMixin.local_project

    @classmethod
    def callback(cls, ctx, param, value) -> set[Path]:
        """
        Use this for click.option / argument callbacks.
        """
        project = ctx.params.get("project")
        return cls(value, project=project).filtered_paths

    @property
    def filtered_paths(self) -> set[Path]:
        """
        Get the filtered set of paths.
        """
        value = self.value
        contract_paths: Iterable[Path]

        if value and isinstance(value, (Path, str)):
            # Given single path.
            contract_paths = (Path(value),)
        elif not value or value == "*":
            # Get all file paths in the project.
            return {p for p in self.project.sources.paths}
        elif isinstance(value, Iterable):
            contract_paths = value
        else:
            raise BadArgumentUsage(f"Not a path or iter[Path]: {value}")

        # Convert source IDs or relative paths to absolute paths.
        path_set = self.lookup(contract_paths)

        # Handle missing compilers.
        if self.missing_compilers:
            # Craft a nice message for all missing compilers.
            missing_ext = ", ".join(sorted(self.missing_compilers))
            message = (
                f"Missing compilers for the following file types: '{missing_ext}'. "
                "Possibly, a compiler plugin is not installed or is "
                "installed but not loading correctly."
            )
            if ".vy" in self.missing_compilers:
                message = f"{message} Is 'ape-vyper' installed?"
            if ".sol" in self.missing_compilers:
                message = f"{message} Is 'ape-solidity' installed?"

            logger.warning(message)

        return path_set

    @property
    def exclude_patterns(self) -> set[str]:
        from ape.utils.basemodel import ManagerAccessMixin as access

        return access.config_manager.get_config("compile").exclude or set()

    def do_exclude(self, path: Union[Path, str]) -> bool:
        return self.project.sources.is_excluded(path)

    def compiler_is_unknown(self, path: Union[Path, str]) -> bool:
        from ape.utils.basemodel import ManagerAccessMixin
        from ape.utils.os import get_full_extension

        ext = get_full_extension(path)
        unknown_compiler = (
            ext and ext not in ManagerAccessMixin.compiler_manager.registered_compilers
        )
        if unknown_compiler and ext not in self.missing_compilers:
            self.missing_compilers.add(ext)

        return bool(unknown_compiler)

    def lookup(self, path_iter: Iterable, path_set: Optional[set] = None) -> set[Path]:
        path_set = path_set or set()
        given_paths = [p for p in path_iter]  # Handle iterators w/o losing it.

        for path_id in given_paths:
            path = Path(path_id)
            contracts_folder = self.project.contracts_folder
            if (
                self.project.path / path.name
            ) == contracts_folder or path.name == contracts_folder.name:
                # Was given the path to the contracts folder.
                path_set = path_set.union({p for p in self.project.sources.paths})

            elif (self.project.path / path).is_dir():
                # Was given sub-dir in the project folder.
                path_set = path_set.union(
                    self.lookup(
                        (p for p in (self.project.path / path).iterdir()), path_set=path_set
                    )
                )

            elif (contracts_folder / path.name).is_dir():
                # Was given sub-dir in the contracts folder.
                path_set = path_set.union(
                    self.lookup(
                        (p for p in (contracts_folder / path.name).iterdir()), path_set=path_set
                    )
                )

            elif resolved_path := self.project.sources.lookup(path):
                # Check compiler missing.
                if self.compiler_is_unknown(resolved_path):
                    # NOTE: ^ Also tracks.
                    continue

                # We know here that the compiler is known.
                path_set.add(resolved_path)

            else:
                raise BadArgumentUsage(f"Source file '{path.name}' not found.")

        return path_set


def contract_file_paths_argument():
    """
    A ``click.argument`` representing contract source file paths.
    This argument takes 0-to-many values.

    The return type from the callback is a flattened list of
    source file-paths.
    """
    return click.argument("file_paths", nargs=-1, callback=_ContractPaths.callback)



import pytest
from ethpm_types import ContractType

from ape import Contract
from ape.contracts import ContractInstance
from ape.exceptions import ContractNotFoundError, ConversionError
from ape.logging import LogLevel, logger
from ape.managers._contractscache import _merge_contract_types
from ape_ethereum.proxies import ProxyInfo, ProxyType, _make_minimal_proxy
from tests.conftest import explorer_test, skip_if_plugin_installed


@pytest.fixture
def contract_0(vyper_contract_container):
    return vyper_contract_container


@pytest.fixture
def contract_1(solidity_contract_container):
    return solidity_contract_container


def test_merge_contract_types(contract_instance):
    ct = contract_instance.contract_type
    modified_ct = ct.model_copy(deep=True)
    modified_ct.view_methods[0].name = "foo"
    new_ct = _merge_contract_types(ct, modified_ct)
    assert len(new_ct.abi) == len(ct.abi) + 1 == len(modified_ct.abi) + 1
    assert len(new_ct.abi) == len(_merge_contract_types(new_ct, modified_ct).abi)


def test_instance_at(chain, contract_instance):
    contract = chain.contracts.instance_at(str(contract_instance.address))
    assert contract.contract_type == contract_instance.contract_type


def test_instance_at_unknown_hex_str(chain, contract_instance):
    # Fails when decoding Ethereum address and NOT conversion error.
    hex_str = "0x1402b10CA274cD76C441e16C844223F79D3566De12bb12b0aebFE41aDFAe302"
    with pytest.raises(ValueError, match=f"Unknown address value '{hex_str}'."):
        chain.contracts.instance_at(hex_str)


def test_instance_at_when_given_contract_type(chain, contract_instance):
    contract = chain.contracts.instance_at(
        str(contract_instance.address), contract_type=contract_instance.contract_type
    )
    assert contract.contract_type == contract_instance.contract_type


def test_instance_at_when_given_name_as_contract_type(chain, contract_instance):
    expected_match = "Expected type 'ContractType' for argument 'contract_type'; Given 'str'."
    with pytest.raises(TypeError, match=expected_match):
        address = str(contract_instance.address)
        bad_contract_type = contract_instance.contract_type.name
        chain.contracts.instance_at(address, contract_type=bad_contract_type)


@explorer_test
def test_instance_at_uses_given_contract_type_when_retrieval_fails(mocker, chain, ape_caplog):
    # The manager always attempts retrieval so that default contact types can
    # get cached. However, sometimes an explorer plugin may fail. If given a contract-type
    # in that situation, we can use it and not fail and log the error instead.
    expected_contract_type = ContractType(contractName="foo", sourceId="foo.bar")
    new_address = "0x4a986a6dCA6dbf99bC3d17F8D71aFb0d60e740f8"
    expected_fail_message = "LOOK_FOR_THIS_FAIL_MESSAGE"
    existing_fn = chain.contracts.get

    def fn(addr, default=None, **kwargs):
        if addr == new_address:
            raise ValueError(expected_fail_message)

        return existing_fn(addr, default=default)

    original_get = chain.contracts.get
    mock_get = mocker.MagicMock()
    mock_get.side_effect = fn
    chain.contracts.get = mock_get
    try:
        actual = chain.contracts.instance_at(new_address, contract_type=expected_contract_type)
    finally:
        chain.contracts.get = original_get

    ape_caplog.assert_last_log(expected_fail_message)
    assert actual.contract_type == expected_contract_type


@explorer_test
def test_instance_at_contract_type_not_found_local_network(chain, eth_tester_provider):
    eth_tester_provider.network.__dict__["explorer"] = None
    new_address = "0x4a986a6dca6dbF99Bc3D17F8d71aFB0D60E740F9"
    expected = rf"Failed to get contract type for address '{new_address}'."
    with pytest.raises(ContractNotFoundError, match=expected):
        chain.contracts.instance_at(new_address)


@explorer_test
def test_instance_at_contract_type_not_found_live_network(chain, eth_tester_provider):
    eth_tester_provider.network.__dict__["explorer"] = None
    real_name = eth_tester_provider.network.name
    eth_tester_provider.network.name = "sepolia"
    try:
        new_address = "0x4a986a6dca6dbF99Bc3D17F8d71aFB0D60E740F9"
        expected = (
            rf"Failed to get contract type for address '{new_address}'. "
            r"Current network 'ethereum:sepolia:test' has no associated explorer plugin. "
            "Try installing an explorer plugin using .*ape plugins install etherscan.*, "
            r"or using a network with explorer support\."
        )
        with pytest.raises(ContractNotFoundError, match=expected):
            chain.contracts.instance_at(new_address)

    finally:
        eth_tester_provider.network.name = real_name


def test_instance_at_use_abi(chain, solidity_fallback_contract, owner):
    new_instance = owner.deploy(solidity_fallback_contract.contract_type)
    del chain.contracts[new_instance.address]
    with pytest.raises(ContractNotFoundError):
        _ = chain.contracts.instance_at(new_instance.address)

    # Now, use only ABI and ensure it works and caches!
    abi = solidity_fallback_contract.contract_type.abi
    instance = chain.contracts.instance_at(new_instance.address, abi=abi)
    assert instance.contract_type.abi

    # `abi=` not needed this time.
    instance2 = chain.contracts.instance_at(new_instance.address)
    assert instance2.contract_type.abi == instance.contract_type.abi


def test_instance_at_provide_proxy(mocker, chain, vyper_contract_instance, owner):
    address = vyper_contract_instance.address
    container = _make_minimal_proxy(address=address.lower())
    proxy = container.deploy(sender=owner)
    proxy_info = chain.contracts.proxy_infos[proxy.address]

    del chain.contracts[proxy.address]

    proxy_detection_spy = mocker.spy(chain.contracts.proxy_infos, "get_type")

    with pytest.raises(ContractNotFoundError):
        # This just fails because we deleted it from the cache so Ape no
        # longer knows what the contract type is. That is fine for this test!
        chain.contracts.instance_at(proxy.address, proxy_info=proxy_info)

    # The real test: we check the spy to ensure we never attempted to look up
    # the proxy info for the given address to `instance_at()`.
    for call in proxy_detection_spy.call_args_list:
        for arg in call[0]:
            assert proxy.address != arg


def test_instance_at_skip_proxy(mocker, chain, vyper_contract_instance, owner):
    address = vyper_contract_instance.address
    del chain.contracts[address]
    proxy_detection_spy = mocker.spy(chain.contracts.proxy_infos, "get_type")

    with pytest.raises(ContractNotFoundError):
        # This just fails because we deleted it from the cache so Ape no
        # longer knows what the contract type is. That is fine for this test!
        chain.contracts.instance_at(address, detect_proxy=False)

    # The real test: we check the spy to ensure we never attempted to look up
    # the proxy info for the given address to `instance_at()`.
    for call in proxy_detection_spy.call_args_list:
        for arg in call[0]:
            assert address != arg


def test_cache_deployment_live_network(
    chain,
    dummy_live_network,
    clean_contract_caches,
    vyper_contract_instance,
    vyper_contract_container,
):
    # Arrange - Ensure the contract is not cached anywhere
    address = vyper_contract_instance.address
    contract_name = vyper_contract_instance.contract_type.name
    contract_types = chain.contracts.contract_types.memory
    chain.contracts.contract_types.memory = {
        a: ct for a, ct in contract_types.items() if a != address
    }
    del chain.contracts.deployments[contract_name]

    # Act
    chain.contracts.cache_deployment(vyper_contract_instance)

    # Assert
    actual_deployments = chain.contracts.get_deployments(vyper_contract_container)
    actual_contract_type = chain.contracts.contract_types[address]
    expected = vyper_contract_instance.contract_type
    assert len(actual_deployments) == 1
    assert actual_deployments[0].address == address
    assert actual_deployments[0].txn_hash == vyper_contract_instance.txn_hash
    assert chain.contracts.get(address) == expected
    assert actual_contract_type == expected


def test_cache_default_contract_type_when_used(solidity_contract_instance, chain, config):
    address = solidity_contract_instance.address
    contract_type = solidity_contract_instance.contract_type

    # Delete contract from local cache if it's there
    del chain.contracts[address]

    # Create a contract using the contract type when nothing is cached.
    contract = Contract(address, contract_type=contract_type)
    assert isinstance(contract, ContractInstance)

    # Ensure we don't need the contract type when creating it the second time.
    contract = Contract(address)
    assert isinstance(contract, ContractInstance)


@explorer_test
def test_contracts_getitem_contract_not_found(chain, eth_tester_provider):
    eth_tester_provider.network.__dict__["explorer"] = None
    new_address = "0x4a986a6dca6dbF99Bc3D17F8d71aFB0D60E740F9"
    real_name = eth_tester_provider.network.name
    eth_tester_provider.network.name = "sepolia"
    try:
        expected = (
            rf"Failed to get contract type for address '{new_address}'. "
            r"Current network 'ethereum:sepolia:test' has no associated explorer plugin. "
            "Try installing an explorer plugin using .*ape plugins install etherscan.*, "
            r"or using a network with explorer support\."
        )
        with pytest.raises(KeyError, match=expected):
            _ = chain.contracts[new_address]

    finally:
        eth_tester_provider.network.name = real_name


def test_deployments_mapping_cache_location(chain):
    # Arrange / Act
    mapping_location = chain.contracts.deployments.cachefile
    split_mapping_location = str(mapping_location).split("/")

    # Assert
    assert split_mapping_location[-1] == "deployments_map.json"
    assert split_mapping_location[-2] == "ethereum"


def test_deployments_when_offline(chain, networks_disconnected, vyper_contract_container):
    """
    Ensure you don't get `ProviderNotConnectedError` here.
    """
    assert chain.contracts.get_deployments(vyper_contract_container) == []


def test_get_deployments_local(chain, owner, contract_0, contract_1):
    # Arrange
    chain.contracts.clear_local_caches()
    deployed_contract_0 = owner.deploy(contract_0, 900000000)
    deployed_contract_1 = owner.deploy(contract_1, 900000001)

    # Act
    contracts_list_0 = chain.contracts.get_deployments(contract_0)
    contracts_list_1 = chain.contracts.get_deployments(contract_1)

    # Assert
    for contract_list in (contracts_list_0, contracts_list_1):
        assert type(contract_list[0]) is ContractInstance

    actual_address_0 = contracts_list_0[-1].address
    assert actual_address_0 == deployed_contract_0.address
    actual_address_1 = contracts_list_1[-1].address
    assert actual_address_1 == deployed_contract_1.address


def test_get_deployments_live(
    chain, owner, contract_0, contract_1, clean_contract_caches, dummy_live_network
):
    deployed_contract_0 = owner.deploy(contract_0, 8000000, required_confirmations=0)
    deployed_contract_1 = owner.deploy(contract_1, 8000001, required_confirmations=0)

    # Act
    my_contracts_list_0 = chain.contracts.get_deployments(contract_0)
    my_contracts_list_1 = chain.contracts.get_deployments(contract_1)

    # Assert
    address_from_api_0 = my_contracts_list_0[-1].address
    assert address_from_api_0 == deployed_contract_0.address
    address_from_api_1 = my_contracts_list_1[-1].address
    assert address_from_api_1 == deployed_contract_1.address


def test_get_multiple_deployments_live(
    chain, owner, contract_0, contract_1, clean_contract_caches, dummy_live_network
):
    starting_contracts_list_0 = chain.contracts.get_deployments(contract_0)
    starting_contracts_list_1 = chain.contracts.get_deployments(contract_1)
    initial_deployed_contract_0 = owner.deploy(contract_0, 700000, required_confirmations=0)
    initial_deployed_contract_1 = owner.deploy(contract_1, 700001, required_confirmations=0)
    owner.deploy(contract_0, 700002, required_confirmations=0)
    owner.deploy(contract_1, 700003, required_confirmations=0)
    final_deployed_contract_0 = owner.deploy(contract_0, 600000, required_confirmations=0)
    final_deployed_contract_1 = owner.deploy(contract_1, 600001, required_confirmations=0)
    contracts_list_0 = chain.contracts.get_deployments(contract_0)
    contracts_list_1 = chain.contracts.get_deployments(contract_1)
    contract_type_map = {
        "ApeContract0": (initial_deployed_contract_0, final_deployed_contract_0),
        "ApeContract1": (initial_deployed_contract_1, final_deployed_contract_1),
    }

    assert len(contracts_list_0) == len(starting_contracts_list_0) + 3
    assert len(contracts_list_1) == len(starting_contracts_list_1) + 3

    for ct_name, ls in zip(("ApeContract0", "ApeContract1"), (contracts_list_0, contracts_list_1)):
        initial_ct, final_ct = contract_type_map[ct_name]
        assert ls[len(ls) - 3].address == initial_ct.address
        assert ls[-1].address == final_ct.address


def test_cache_updates_per_deploy(owner, chain, contract_0, contract_1):
    # Arrange / Act
    initial_contracts = chain.contracts.get_deployments(contract_0)
    expected_first_contract = owner.deploy(contract_0, 6787678)

    owner.deploy(contract_0, 6787679)
    owner.deploy(contract_0, 6787680)
    expected_last_contract = owner.deploy(contract_0, 6787681)

    actual_contracts = chain.contracts.get_deployments(contract_0)
    first_index = len(initial_contracts)  # next index before deploys from this test
    actual_first_contract = actual_contracts[first_index].address
    actual_last_contract = actual_contracts[-1].address

    # Assert
    fail_msg = f"Check deployments: {', '.join([c.address for c in actual_contracts])}"
    assert len(actual_contracts) - len(initial_contracts) == 4, fail_msg
    assert actual_first_contract == expected_first_contract.address, fail_msg
    assert actual_last_contract == expected_last_contract.address, fail_msg


def test_get_multiple(vyper_contract_instance, solidity_contract_instance, chain):
    contract_map = chain.contracts.get_multiple(
        (vyper_contract_instance.address, solidity_contract_instance.address)
    )
    assert len(contract_map) == 2
    assert contract_map[vyper_contract_instance.address] == vyper_contract_instance.contract_type
    assert (
        contract_map[solidity_contract_instance.address] == solidity_contract_instance.contract_type
    )


def test_get_multiple_no_addresses(chain, ape_caplog):
    with ape_caplog.at_level(LogLevel.WARNING):
        contract_map = chain.contracts.get_multiple([])
        assert "WARNING" in ape_caplog.records[-1].levelname
        assert "No addresses provided." in ape_caplog.messages[-1]

    assert not contract_map


def test_get_multiple_include_non_contract_address(vyper_contract_instance, chain, owner):
    actual = chain.contracts.get_multiple((vyper_contract_instance.address, owner.address))
    assert len(actual) == 1
    assert actual[vyper_contract_instance.address] == vyper_contract_instance.contract_type


@skip_if_plugin_installed("ens")
def test_get_multiple_attempts_to_convert(chain):
    with pytest.raises(ConversionError):
        chain.contracts.get_multiple(("test.eth",))


def test_get_non_contract_address(chain, owner):
    actual = chain.contracts.get(owner.address)
    assert actual is None


def test_get_attempts_to_convert(chain):
    with pytest.raises(ConversionError):
        # NOTE: using eth2 suffix so still works if ape-ens is installed.
        chain.contracts.get("test.eth2")


@explorer_test
def test_get_attempts_explorer(
    mock_explorer, create_mock_sepolia, chain, owner, vyper_fallback_container
):
    contract = owner.deploy(vyper_fallback_container)

    def get_contract_type(addr):
        if addr == contract.address:
            return contract.contract_type

        raise ValueError("nope")

    # Hack in a way to publish on this local network.
    with create_mock_sepolia() as network:
        del chain.contracts[contract.address]
        mock_explorer.get_contract_type.side_effect = get_contract_type
        network.__dict__["explorer"] = mock_explorer
        try:
            actual = chain.contracts.get(contract.address)
        finally:
            network.__dict__["explorer"] = None

        assert actual == contract.contract_type
        assert mock_explorer.get_contract_type.call_count > 0
        mock_explorer.get_contract_type.reset_mock()


@explorer_test
def test_get_attempts_explorer_logs_errors_from_explorer(
    mock_explorer, create_mock_sepolia, chain, owner, vyper_fallback_container, ape_caplog
):
    contract = owner.deploy(vyper_fallback_container)
    check_error_str = "__CHECK_FOR_THIS_ERROR__"
    expected_log = (
        f"Attempted to retrieve contract type from explorer 'mock' "
        f"from address '{contract.address}' but encountered an "
        f"exception: {check_error_str}"
    )

    def get_contract_type(addr):
        if addr == contract.address:
            raise ValueError(check_error_str)

        raise ValueError("nope")

    with create_mock_sepolia() as network:
        del chain.contracts[contract.address]
        mock_explorer.get_contract_type.side_effect = get_contract_type
        network.__dict__["explorer"] = mock_explorer
        try:
            actual = chain.contracts.get(contract.address)
        finally:
            network.__dict__["explorer"] = None

        assert expected_log in ape_caplog.head
        assert actual is None
        mock_explorer.get_contract_type.reset_mock()


@explorer_test
def test_get_attempts_explorer_logs_rate_limit_error_from_explorer(
    mock_explorer, create_mock_sepolia, chain, owner, vyper_fallback_container, ape_caplog
):
    contract = owner.deploy(vyper_fallback_container)

    # For rate limit errors, we don't show anything else,
    # as it may be confusing.
    check_error_str = "you have been rate limited"

    def get_contract_type(addr):
        if addr == contract.address:
            raise ValueError(check_error_str)

        raise ValueError("nope")

    with create_mock_sepolia() as network:
        # Ensure is not cached locally.
        del chain.contracts[contract.address]

        mock_explorer.get_contract_type.side_effect = get_contract_type
        network.__dict__["explorer"] = mock_explorer
        try:
            with logger.at_level(LogLevel.INFO):
                actual = chain.contracts.get(contract.address)
        finally:
            network.__dict__["explorer"] = None

        assert check_error_str in ape_caplog.head
        assert actual is None
        mock_explorer.get_contract_type.reset_mock()


def test_cache_non_checksum_address(chain, vyper_contract_instance):
    """
    When caching a non-checksum address, it should use its checksum
    form automatically.
    """
    if vyper_contract_instance.address in chain.contracts:
        del chain.contracts[vyper_contract_instance.address]

    lowered_address = vyper_contract_instance.address.lower()
    chain.contracts[lowered_address] = vyper_contract_instance.contract_type
    assert chain.contracts[vyper_contract_instance.address] == vyper_contract_instance.contract_type


def test_get_proxy(chain, owner, minimal_proxy_container, vyper_contract_instance):
    placeholder = "0xBEbeBeBEbeBebeBeBEBEbebEBeBeBebeBeBebebe"
    if placeholder in chain.contracts:
        del chain.contracts[placeholder]

    minimal_proxy = owner.deploy(minimal_proxy_container, sender=owner)
    chain.provider.network.__dict__["explorer"] = None  # Ensure no explorer, messes up test.

    actual = chain.contracts.get(minimal_proxy.address)
    assert actual == minimal_proxy.contract_type


def test_get_proxy_implementation_missing(chain, owner, vyper_contract_container):
    """
    Proxy is cached but implementation is missing.
    """
    placeholder = vyper_contract_container.deploy(1001, sender=owner)
    assert chain.contracts[placeholder.address]  # This must be cached!

    proxy_container = _make_minimal_proxy(placeholder.address)
    minimal_proxy = owner.deploy(proxy_container, sender=owner)
    chain.provider.network.__dict__["explorer"] = None  # Ensure no explorer, messes up test.

    if minimal_proxy.address in chain.contracts:
        # Delete the proxy but make sure it does not delete the implementation!
        # (which it normally does here).
        del chain.contracts[minimal_proxy.address]
        chain.contracts[placeholder.address] = placeholder

    actual = chain.contracts.get(minimal_proxy.address)
    assert actual == minimal_proxy.contract_type


def test_get_proxy_pass_proxy_info(chain, owner, minimal_proxy_container, ethereum):
    placeholder = "0xBEbeBeBEbeBebeBeBEBEbebEBeBeBebeBeBebebe"
    if placeholder in chain.contracts:
        del chain.contracts[placeholder]

    minimal_proxy = owner.deploy(minimal_proxy_container, sender=owner)
    chain.provider.network.__dict__["explorer"] = None  # Ensure no explorer, messes up test.
    info = ethereum.get_proxy_info(minimal_proxy.address)
    assert info

    # Ensure not already cached.
    if minimal_proxy.address in chain.contracts:
        del chain.contracts[minimal_proxy.address]

    actual = chain.contracts.get(minimal_proxy.address, proxy_info=info)
    assert actual is None  # It can't find the contact anymore.

    # Ensure it does store 'None' (was a bug where it did).
    assert minimal_proxy.address not in chain.contracts.contract_types


@explorer_test
def test_get_proxy_pass_proxy_info_and_no_explorer(
    chain, owner, proxy_contract_container, ethereum, dummy_live_network_with_explorer
):
    """
    Tests the condition of both passing `proxy_info=` and setting `use_explorer=False`
    when getting the ContractType of a proxy.
    """
    explorer = dummy_live_network_with_explorer.explorer
    placeholder = "0xBEbeBeBEbeBebeBeBEBEbebEBeBeBebeBeBebebe"
    if placeholder in chain.contracts:
        del chain.contracts[placeholder]

    proxy = proxy_contract_container.deploy(placeholder, sender=owner, required_confirmations=0)
    info = ProxyInfo(type=ProxyType.Minimal, target=placeholder)
    explorer.get_contract_type.reset_mock()
    chain.contracts.get(proxy.address, proxy_info=info, fetch_from_explorer=False)

    # Ensure explorer was not used.
    assert explorer.get_contract_type.call_count == 0


def test_get_creation_metadata(chain, vyper_contract_instance, owner):
    address = vyper_contract_instance.address
    creation = chain.contracts.get_creation_metadata(address)
    assert creation.deployer == owner.address

    chain.mine()
    creation = chain.contracts.get_creation_metadata(address)
    assert creation.deployer == owner.address


def test_delete_contract(vyper_contract_instance, chain):
    # Ensure we start with it cached.
    chain.contracts[vyper_contract_instance.address] = vyper_contract_instance

    del chain.contracts[vyper_contract_instance.address]
    assert vyper_contract_instance.address not in chain.contracts

    # Ensure we can't access it.
    with pytest.raises(KeyError):
        _ = chain.contracts[vyper_contract_instance.address]


def test_delete_proxy(vyper_contract_instance, chain, ethereum, owner):
    address = vyper_contract_instance.address
    container = _make_minimal_proxy(address=address.lower())
    proxy = container.deploy(sender=owner)

    # Ensure we start with both the proxy and the target contracts cached.
    if proxy.address not in chain.contracts:
        chain.contracts[proxy.address] = proxy

    proxy_info = ethereum.get_proxy_info(proxy.address)
    chain.contracts.cache_proxy_info(proxy.address, proxy_info)
    if proxy_info.target not in chain.contracts:
        chain.contracts[proxy_info.target] = vyper_contract_instance

    del chain.contracts[proxy.address]
    assert proxy.address not in chain.contracts

    # Ensure we can't access it.
    with pytest.raises(KeyError):
        _ = chain.contracts[proxy.address]

    # Ensure we can't access the target either.
    with pytest.raises(KeyError):
        _ = chain.contracts[proxy_info.target]


def test_clear_local_caches(chain, vyper_contract_instance, proxy_contract_container, owner):
    # Ensure contract type exists.
    address = vyper_contract_instance.address
    # Ensure blueprint exists.
    chain.contracts.blueprints[address] = vyper_contract_instance.contract_type
    # Ensure proxy exists.
    proxy = proxy_contract_container.deploy(address, sender=owner)
    # Ensure creation exists.
    _ = chain.contracts.get_creation_metadata(address)

    # Test setup verification.
    assert address in chain.contracts.contract_types, "Setup failed - no contract type(s) cached"
    assert proxy.address in chain.contracts.proxy_infos, "Setup failed - no proxy cached"
    assert address in chain.contracts.contract_creations, "Setup failed - no creation(s) cached"

    # This is the method we are testing.
    chain.contracts.clear_local_caches()

    # Assertions - everything should be empty.
    assert chain.contracts.proxy_infos.memory == {}
    assert chain.contracts.blueprints.memory == {}
    assert chain.contracts.contract_types.memory == {}
    assert chain.contracts.contract_creations.memory == {}



import os
import shutil
from collections.abc import Iterable, Iterator
from contextlib import contextmanager
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union

from ape.api.config import ApeConfig
from ape.managers.base import BaseManager
from ape.utils.basemodel import (
    ExtraAttributesMixin,
    ExtraModelAttributes,
    get_attribute_with_extras,
    get_item_with_extras,
    only_raise_attribute_error,
)
from ape.utils.misc import log_instead_of_fail
from ape.utils.os import create_tempdir, in_tempdir
from ape.utils.rpc import USER_AGENT, RPCHeaders

if TYPE_CHECKING:
    from ethpm_types import PackageManifest


CONFIG_FILE_NAME = "ape-config.yaml"


class ConfigManager(ExtraAttributesMixin, BaseManager):
    """
    An Ape configuration manager, controlled by ``ape-config.yaml``
    files. **NOTE**: This is a singleton wrapper class that
    points to the local project's config. For the config field
    definitions, see :class:`~ape.api.config.ApeConfig`.
    """

    def __init__(self, data_folder: Optional[Path] = None, request_header: Optional[dict] = None):
        if not data_folder and "APE_DATA_FOLDER" in os.environ:
            self.DATA_FOLDER = Path(os.environ["APE_DATA_FOLDER"])
        else:
            self.DATA_FOLDER = data_folder or Path.home() / ".ape"

        request_header = request_header or {
            "User-Agent": USER_AGENT,
            "Content-Type": "application/json",
        }
        self.REQUEST_HEADER = request_header or {}

    def __ape_extra_attributes__(self):
        # The "extra" attributes are the local project's
        # config attributes. To see the actual ``ape-config.yaml``
        # definitions, see :class:`~ape.api.config.ApeConfig`.
        yield ExtraModelAttributes(
            name="config",
            # Active project's config.
            attributes=self.local_project.config,
            include_getitem=True,
        )

    @log_instead_of_fail(default="<ConfigManager>")
    def __repr__(self) -> str:
        return f"<{CONFIG_FILE_NAME}>"

    def __str__(self) -> str:
        return str(self.local_project.config)

    @only_raise_attribute_error
    def __getattr__(self, name: str) -> Any:
        """
        The root config manager (funneling to this method)
        refers to the local project's config. Config is loaded
        per project in Ape to support multi-project environments
        and a smarter dependency system.

        See :class:`~ape.api.config.ApeConfig` for field definitions
        and model-related controls.
        """
        return get_attribute_with_extras(self, name)

    def __getitem__(self, name: str) -> Any:
        return get_item_with_extras(self, name)

    @cached_property
    def global_config(self) -> ApeConfig:
        """
        Root-level configurations, loaded from the
        data folder. **NOTE**: This only needs to load
        once and applies to all projects.
        """
        return self.load_global_config()

    def get_config(self, name: str) -> ApeConfig:
        return self.local_project.config.get_config(name)

    def load_global_config(self) -> ApeConfig:
        path = self.DATA_FOLDER / CONFIG_FILE_NAME
        return ApeConfig.validate_file(path) if path.is_file() else ApeConfig.model_validate({})

    def merge_with_global(self, project_config: ApeConfig) -> ApeConfig:
        global_data = self.global_config.model_dump(by_alias=True)
        project_data = project_config.model_dump(by_alias=True)
        merged_data = merge_configs(global_data, project_data)
        return ApeConfig.model_validate(merged_data)

    @classmethod
    def extract_config(cls, manifest: "PackageManifest", **overrides) -> ApeConfig:
        """
        Calculate the ape-config data from a package manifest.

        Args:
            manifest (PackageManifest): The manifest.
            **overrides: Custom config settings.

        Returns:
            :class:`~ape.managers.config.ApeConfig`: Config data.
        """
        return ApeConfig.from_manifest(manifest, **overrides)

    @contextmanager
    def isolate_data_folder(
        self, keep: Optional[Union[Iterable[str], str]] = None
    ) -> Iterator[Path]:
        """
        Change Ape's DATA_FOLDER to point a temporary path,
        in a context, for testing purposes. Any data
        cached to disk will not persist.

        Args:
            keep (Optional[Union[Iterable[str], str]]): Optionally, pass in
              a key of subdirectory names to include in the new isolated
              data folder. For example, pass ing ``"packages"`` to avoid
              having to re-download dependencies in an isolated environment.

        Returns:
            Iterator[Path]: The temporary data folder.
        """
        original_data_folder = self.DATA_FOLDER
        if in_tempdir(original_data_folder):
            # Already isolated.
            yield original_data_folder

        else:
            keep = [keep] if isinstance(keep, str) else keep or []
            try:
                with create_tempdir() as temp_data_folder:
                    # Copy in items from "keep".
                    for item in keep:
                        path_to_keep = original_data_folder / item
                        if not path_to_keep.is_dir():
                            continue

                        dest_path = temp_data_folder / item
                        shutil.copytree(path_to_keep, dest_path)

                    self.DATA_FOLDER = temp_data_folder
                    yield temp_data_folder

            finally:
                self.DATA_FOLDER = original_data_folder

    def _get_request_headers(self) -> RPCHeaders:
        # Avoid multiple keys error by not initializing with both dicts.
        headers = RPCHeaders(**self.REQUEST_HEADER)
        for key, value in self.request_headers.items():
            headers[key] = value

        return headers


def merge_configs(*cfgs: dict) -> dict:
    if len(cfgs) == 0:
        return {}
    elif len(cfgs) == 1:
        return cfgs[0]

    new_base = _merge_configs(cfgs[0], cfgs[1])
    return merge_configs(new_base, *cfgs[2:])


def _merge_configs(base: dict, secondary: dict) -> dict:
    result: dict = {}

    # Short circuits
    if not base and not secondary:
        return result
    elif not base:
        return secondary
    elif not secondary:
        return base

    for key, value in base.items():
        if key not in secondary:
            result[key] = value

        elif not isinstance(value, dict) or not isinstance(secondary[key], dict):
            # Is a primitive value found in both configs.
            # Must use the second one.
            result[key] = secondary[key]

        else:
            # Merge the dictionaries.
            sub = _merge_configs(value, secondary[key])
            result[key] = sub

    # Add missed keys from secondary.
    for key, value in secondary.items():
        if key not in base:
            result[key] = value

    return result



import os
from abc import abstractmethod
from collections.abc import Iterator
from contextlib import contextmanager
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union

import click
from eip712.messages import EIP712Message
from eip712.messages import SignableMessage as EIP712SignableMessage
from eth_account import Account
from eth_account.messages import encode_defunct
from eth_utils import to_hex
from ethpm_types import ContractType

from ape.api.address import BaseAddress
from ape.api.transactions import ReceiptAPI, TransactionAPI
from ape.exceptions import (
    AccountsError,
    AliasAlreadyInUseError,
    APINotImplementedError,
    ConversionError,
    MethodNonPayableError,
    MissingDeploymentBytecodeError,
    SignatureError,
    TransactionError,
)
from ape.logging import logger
from ape.types.address import AddressType
from ape.types.signatures import MessageSignature, SignableMessage
from ape.utils.basemodel import BaseInterfaceModel
from ape.utils.misc import raises_not_implemented
from ape.utils.testing import (
    DEFAULT_NUMBER_OF_TEST_ACCOUNTS,
    DEFAULT_TEST_HD_PATH,
    DEFAULT_TEST_MNEMONIC,
)

if TYPE_CHECKING:
    from eth_pydantic_types import HexBytes

    from ape.contracts import ContractContainer, ContractInstance


class AccountAPI(BaseInterfaceModel, BaseAddress):
    """
    An API class representing an account.
    """

    def __dir__(self) -> list[str]:
        """
        Display methods to IPython on ``a.[TAB]`` tab completion.

        Returns:
            list[str]: Method names that IPython uses for tab completion.
        """
        base_value_excludes = ("code", "codesize", "is_contract")  # Not needed for accounts
        base_values = [v for v in self._base_dir_values if v not in base_value_excludes]
        return base_values + [
            self.__class__.alias.fget.__name__,  # type: ignore[attr-defined]
            self.__class__.call.__name__,
            self.__class__.deploy.__name__,
            self.__class__.prepare_transaction.__name__,
            self.__class__.sign_authorization.__name__,
            self.__class__.sign_message.__name__,
            self.__class__.sign_transaction.__name__,
            self.__class__.transfer.__name__,
            self.__class__.delegate.fget.__name__,  # type: ignore[attr-defined]
            self.__class__.set_delegate.__name__,
            self.__class__.remove_delegate.__name__,
            self.__class__.delegate_to.__name__,
        ]

    @property
    def alias(self) -> Optional[str]:
        """
        A shortened-name for quicker access to the account.
        """
        return None

    @property
    def public_key(self) -> Optional["HexBytes"]:
        """
        The public key for the account.

        ```{notice}
        Account might not have this property if feature is unsupported or inaccessible.
        ```
        """
        return None

    def prepare_transaction(self, txn: "TransactionAPI", **kwargs) -> "TransactionAPI":
        sign = kwargs.pop("sign", False)
        prepared_tx = super().prepare_transaction(txn, **kwargs)
        return (self.sign_transaction(prepared_tx) or prepared_tx) if sign else prepared_tx

    def sign_raw_msghash(self, msghash: "HexBytes") -> Optional[MessageSignature]:
        """
        Sign a raw message hash.

        Args:
          msghash (:class:`~eth_pydantic_types.HexBytes`):
            The message hash to sign. Plugins may or may not support this operation.
            Default implementation is to raise ``APINotImplementedError``.

        Returns:
          :class:`~ape.types.signatures.MessageSignature` (optional):
            The signature corresponding to the message.
        """
        raise APINotImplementedError(
            f"Raw message signing is not supported by '{self.__class__.__name__}'"
        )

    def sign_authorization(
        self,
        address: Any,
        chain_id: Optional[int] = None,
        nonce: Optional[int] = None,
    ) -> Optional[MessageSignature]:
        """
        Sign an `EIP-7702 <https://eips.ethereum.org/EIPS/eip-7702>`__ Authorization.

        Args:
          address (Any): A delegate address to sign the authorization for.
          chain_id (Optional[int]):
            The chain ID that the authorization should be valid for.
            A value of ``0`` means that the authorization is valid for **any chain**.
            Default tells implementation to use the currently connected network's ``chain_id``.
          nonce (Optional[int]):
            The nonce to use to sign authorization with. Defaults to account's current nonce.

        Returns:
          :class:`~ape.types.signatures.MessageSignature` (optional):
            The signature corresponding to the message.

        ```{caution}
        This action has the capability to be extremely destructive to the signer, and might lead to
        full account compromise. All implementations are recommended to ensure that the signer be
        made aware of the severity and impact of this action through some callout.
        ```
        """

        raise APINotImplementedError(
            f"Authorization signing is not supported by '{self.__class__.__name__}'"
        )

    @abstractmethod
    def sign_message(self, msg: Any, **signer_options) -> Optional[MessageSignature]:
        """
        Sign a message.

        Args:
          msg (Any): The message to sign. Account plugins can handle various types of messages.
            For example, :class:`~ape_accounts.accounts.KeyfileAccount` can handle
            :class:`~ape.types.signatures.SignableMessage`, str, int, and bytes.
            See these
            `docs <https://eth-account.readthedocs.io/en/stable/eth_account.html#eth_account.messages.SignableMessage>`__  # noqa: E501
            for more type information on the :class:`~ape.types.signatures.SignableMessage` type.
          **signer_options: Additional kwargs given to the signer to modify the signing operation.

        Returns:
          :class:`~ape.types.signatures.MessageSignature` (optional): The signature corresponding to the message.
        """

    @abstractmethod
    def sign_transaction(self, txn: TransactionAPI, **signer_options) -> Optional[TransactionAPI]:
        """
        Sign a transaction.

        Args:
          txn (:class:`~ape.api.transactions.TransactionAPI`): The transaction to sign.
          **signer_options: Additional kwargs given to the signer to modify the signing operation.

        Returns:
          :class:`~ape.api.transactions.TransactionAPI` (optional): A signed transaction.
            The ``TransactionAPI`` returned by this method may not correspond to ``txn`` given as
            input, however returning a properly-formatted transaction here is meant to be executed.
            Returns ``None`` if the account does not have a transaction it wishes to execute.

        """

    def call(
        self,
        txn: TransactionAPI,
        send_everything: bool = False,
        private: bool = False,
        sign: bool = True,
        **signer_options,
    ) -> ReceiptAPI:
        """
        Make a transaction call.

        Raises:
            :class:`~ape.exceptions.AccountsError`: When the nonce is invalid or the sender does
              not have enough funds.
            :class:`~ape.exceptions.TransactionError`: When the required confirmations are negative.
            :class:`~ape.exceptions.SignatureError`: When the user does not sign the transaction.
            :class:`~ape.exceptions.APINotImplementedError`: When setting ``private=True`` and using
              a provider that does not support private transactions.

        Args:
            txn (:class:`~ape.api.transactions.TransactionAPI`): An invoke-transaction.
            send_everything (bool): ``True`` will send the difference from balance and fee.
              Defaults to ``False``.
            private (bool): ``True`` will use the
              :meth:`~ape.api.providers.ProviderAPI.send_private_transaction` method.
            sign (bool): ``False`` to not sign the transaction (useful for providers like Titanoboa
              which still use a sender but don't need to sign).
            **signer_options: Additional kwargs given to the signer to modify the signing operation.

        Returns:
            :class:`~ape.api.transactions.ReceiptAPI`
        """

        txn = self.prepare_transaction(txn)
        max_fee = txn.max_fee
        gas_limit = txn.gas_limit

        if not isinstance(gas_limit, int):
            raise TransactionError("Transaction not prepared.")

        # The conditions below should never reached but are here for mypy's sake.
        # The `max_fee` was either set manually or from `prepare_transaction()`.
        # The `gas_limit` was either set manually or from `prepare_transaction()`.
        if max_fee is None:
            raise TransactionError("`max_fee` failed to get set in transaction preparation.")
        elif gas_limit is None:
            raise TransactionError("`gas_limit` failed to get set in transaction preparation.")

        total_fees = max_fee * gas_limit

        # Send the whole balance.
        if send_everything:
            amount_to_send = self.balance - total_fees
            if amount_to_send <= 0:
                raise AccountsError(
                    f"Sender does not have enough to cover transaction value and gas: {total_fees}"
                )
            else:
                txn.value = amount_to_send

        if sign:
            prepared_txn = self.sign_transaction(txn, **signer_options)
            if not prepared_txn:
                raise SignatureError("The transaction was not signed.", transaction=txn)

        else:
            prepared_txn = txn

        if not prepared_txn.sender:
            prepared_txn.sender = self.address

        return (
            self.provider.send_private_transaction(prepared_txn)
            if private
            else self.provider.send_transaction(prepared_txn)
        )

    def transfer(
        self,
        account: Union[str, AddressType, BaseAddress],
        value: Optional[Union[str, int]] = None,
        data: Optional[Union[bytes, str]] = None,
        private: bool = False,
        **kwargs,
    ) -> ReceiptAPI:
        """
        Send funds to an account.

        Raises:
            :class:`~ape.exceptions.APINotImplementedError`: When setting ``private=True``
              and using a provider that does not support private transactions.

        Args:
            account (Union[str, AddressType, BaseAddress]): The receiver of the funds.
            value (Optional[Union[str, int]]): The amount to send.
            data (Optional[Union[bytes, str]]): Extra data to include in the transaction.
            private (bool): ``True`` asks the provider to make the transaction
              private. For example, EVM providers typically use the RPC
              ``eth_sendPrivateTransaction`` to achieve this. Local providers may ignore
              this value.
            **kwargs: Additional transaction kwargs passed to
              :meth:`~ape.api.networks.EcosystemAPI.create_transaction`, such as ``gas``
              ``max_fee``, or ``max_priority_fee``. For a list of available transaction
              kwargs, see :class:`~ape.api.transactions.TransactionAPI`.

        Returns:
            :class:`~ape.api.transactions.ReceiptAPI`
        """
        if isinstance(account, int):
            raise AccountsError(
                "Cannot use integer-type for the `receiver` argument in the "
                "`.transfer()` method (this protects against accidentally passing "
                "the `value` as the `receiver`)."
            )

        try:
            receiver = self.conversion_manager.convert(account, AddressType)
        except ConversionError as err:
            raise AccountsError(f"Invalid `receiver` value: '{account}'.") from err

        txn = self.provider.network.ecosystem.create_transaction(
            sender=self.address, receiver=receiver, **kwargs
        )

        if data:
            txn.data = self.conversion_manager.convert(data, bytes)

        if value is None and not kwargs.get("send_everything"):
            raise AccountsError("Must provide 'VALUE' or use 'send_everything=True'")

        elif value is not None and kwargs.get("send_everything"):
            raise AccountsError("Cannot use 'send_everything=True' with 'VALUE'.")

        elif value is not None:
            txn.value = self.conversion_manager.convert(value, int)
            if txn.value < 0:
                raise AccountsError("Value cannot be negative.")

        return self.call(txn, private=private, **kwargs)

    def deploy(
        self, contract: "ContractContainer", *args, publish: bool = False, **kwargs
    ) -> "ContractInstance":
        """
        Create a smart contract on the blockchain. The smart contract must compile before
        deploying and a provider must be active.

        Args:
            contract (:class:`~ape.contracts.base.ContractContainer`): The type of contract to
              deploy.
            publish (bool): Set to ``True`` to attempt explorer contract verification.
              Defaults to ``False``.

        Returns:
            :class:`~ape.contracts.ContractInstance`: An instance of the deployed contract.
        """
        from ape.contracts import ContractContainer

        if isinstance(contract, ContractType):
            # Hack to allow deploying ContractTypes w/o being
            # wrapped in a container first.
            contract = ContractContainer(contract)

        # NOTE: It is important to type check here to prevent cases where user
        #    may accidentally pass in a ContractInstance, which has a very
        #    different implementation for __call__ than ContractContainer.
        elif not isinstance(contract, ContractContainer):
            raise TypeError(
                "contract argument must be a ContractContainer type, "
                "such as 'project.MyContract' where 'MyContract' is the name of "
                "a contract in your project."
            )

        bytecode = contract.contract_type.deployment_bytecode
        if not bytecode or bytecode.bytecode in (None, "", "0x"):
            raise MissingDeploymentBytecodeError(contract.contract_type)

        txn = contract(*args, **kwargs)
        if kwargs.get("value") and not contract.contract_type.constructor.is_payable:
            raise MethodNonPayableError("Sending funds to a non-payable constructor.")

        txn.sender = self.address
        receipt = contract._cache_wrap(lambda: self.call(txn, **kwargs))
        if not (address := receipt.contract_address):
            raise AccountsError(f"'{receipt.txn_hash}' did not create a contract.")

        contract_type = contract.contract_type
        styled_address = click.style(receipt.contract_address, bold=True)
        contract_name = contract_type.name or "<Unnamed Contract>"
        logger.success(f"Contract '{contract_name}' deployed to: {styled_address}")
        instance = self.chain_manager.contracts.instance_from_receipt(receipt, contract_type)
        self.chain_manager.contracts.cache_deployment(instance)

        if publish:
            self.local_project.deployments.track(instance)
            self.provider.network.publish_contract(address)

        instance.base_path = contract.base_path or self.local_project.path
        return instance

    def declare(self, contract: "ContractContainer", *args, **kwargs) -> ReceiptAPI:
        """
        Deploy the "blueprint" of a contract type. For EVM providers, this likely means
        using `EIP-5202 <https://eips.ethereum.org/EIPS/eip-5202>`__, which is implemented
        in the core ``ape-ethereum`` plugin.

        Args:
            contract (:class:`~ape.contracts.base.ContractContainer`): The contract container
              to declare.

        Returns:
            :class:`~ape.api.transactions.ReceiptAPI`: The receipt of the declare transaction.
        """

        transaction = self.provider.network.ecosystem.encode_contract_blueprint(
            contract.contract_type, *args, **kwargs
        )
        receipt = self.call(transaction)
        if receipt.contract_address:
            self.chain_manager.contracts.cache_blueprint(
                receipt.contract_address, contract.contract_type
            )
        else:
            logger.debug("Failed to cache contract declaration: missing contract address.")

        return receipt

    def check_signature(
        self,
        data: Union[SignableMessage, TransactionAPI, str, EIP712Message, int, bytes],
        signature: Optional[MessageSignature] = None,  # TransactionAPI doesn't need it
        recover_using_eip191: bool = True,
    ) -> bool:
        """
        Verify a message or transaction was signed by this account.

        Args:
            data (Union[:class:`~ape.types.signatures.SignableMessage`, :class:`~ape.api.transactions.TransactionAPI`]):  # noqa: E501
              The message or transaction to verify.
            signature (Optional[:class:`~ape.types.signatures.MessageSignature`]):
              The signature to check. Defaults to ``None`` and is not needed when the first
              argument is a transaction class.
            recover_using_eip191 (bool):
              Perform recovery using EIP-191 signed message check. If set False, then will attempt
              recovery as raw hash. `data`` must be a 32 byte hash if this is set False.
              Defaults to ``True``.

        Returns:
            bool: ``True`` if the data was signed by this account. ``False`` otherwise.
        """
        if isinstance(data, str):
            data = encode_defunct(text=data)
        elif isinstance(data, int):
            data = encode_defunct(hexstr=to_hex(data))
        elif isinstance(data, bytes) and (len(data) != 32 or recover_using_eip191):
            data = encode_defunct(data)
        elif isinstance(data, EIP712Message):
            data = data.signable_message
        if isinstance(data, (SignableMessage, EIP712SignableMessage)):
            if signature:
                return self.address == Account.recover_message(data, vrs=signature)

            else:
                raise AccountsError(
                    "Parameter 'signature' required when verifying a 'SignableMessage'."
                )

        elif isinstance(data, TransactionAPI):
            return self.address == Account.recover_transaction(data.serialize_transaction())

        elif isinstance(data, bytes) and len(data) == 32 and not recover_using_eip191:
            return self.address == Account._recover_hash(data, vrs=signature)

        else:
            raise AccountsError(f"Unsupported message type: {type(data)}.")

    def get_deployment_address(self, nonce: Optional[int] = None) -> AddressType:
        """
        Get a contract address before it is deployed. This is useful
        when you need to pass the contract address to another contract
        before deploying it.

        Args:
            nonce (int | None): Optionally provide a nonce. Defaults
              the account's current nonce.

        Returns:
            AddressType: The contract address.
        """
        # Use the connected network, if available. Else, default to Ethereum.
        ecosystem = (
            self.network_manager.active_provider.network.ecosystem
            if self.network_manager.active_provider
            else self.network_manager.ethereum
        )
        nonce = self.nonce if nonce is None else nonce
        return ecosystem.get_deployment_address(self.address, nonce)

    def set_delegate(self, contract: Union[BaseAddress, AddressType, str], **txn_kwargs):
        """
        Have the account class override the value of its ``delegate``. For plugins that support
        this feature, the way they choose to handle it can vary. For example, it could be a call to
        upgrade itself using some built-in method for a smart wallet (with default txn args) e.g.
        the Safe smart wallet (https://github.com/ApeWorX/ape-safe), or it could be to use an EIP-
        7702-like feature available on the network to set a delegate for that account. However if a
        plugin chooses to handle it, the resulting action (if successful) should make sure that the
        value that ``self.delegate`` returns is the same as ``contract`` after it is completed.

        By default, this method raises ``APINotImplementedError`` signaling that support is not
        available for this feature. Calling this may result in other errors if implemented.

        Args:
            contract (`:class:~ape.contracts.ContractInstance`):
                The contract instance to override the delegate with.
            **txn_kwargs: Additional transaction kwargs passed to
              :meth:`~ape.api.networks.EcosystemAPI.create_transaction`, such as ``gas``
              ``max_fee``, or ``max_priority_fee``. For a list of available transaction
              kwargs, see :class:`~ape.api.transactions.TransactionAPI`.
        """
        raise APINotImplementedError

    def remove_delegate(self, **txn_kwargs):
        """
        Has the account class remove the override for the value of its ``delegate``. For plugins
        that support this feature, the way they choose to handle it can vary. For example, on a
        network using an EIP7702-like feature available it will reset the delegate to empty.
        However, if a plugin chooses to handle it, the resulting action (if successful) should
        make sure that the value that ``self.delegate`` returns ``None`` after it is completed.

        By default, this method raises ``APINotImplementedError`` signaling that support is not
        available for this feature. Calling this may result in other errors if implemented.

        Args:
            **txn_kwargs: Additional transaction kwargs passed to
              :meth:`~ape.api.networks.EcosystemAPI.create_transaction`, such as ``gas``
              ``max_fee``, or ``max_priority_fee``. For a list of available transaction
              kwargs, see :class:`~ape.api.transactions.TransactionAPI`.
        """
        raise APINotImplementedError

    @contextmanager
    def delegate_to(
        self,
        new_delegate: Union[BaseAddress, AddressType, str],
        set_txn_kwargs: Optional[dict] = None,
        reset_txn_kwargs: Optional[dict] = None,
        **txn_kwargs,
    ) -> Iterator[BaseAddress]:
        """
        Temporarily override the value of ``delegate`` for the account inside of a context manager,
        and yields a contract instance object whose interface matches that of ``new_delegate``.
        This is useful for ensuring that delegation is only temporarily extended to an account when
        doing a critical action temporarily, such as using an EIP7702 delegate module.

        Args:
            new_delegate (`:class:~ape.contracts.ContractInstance`):
                The contract instance to override the `delegate` with.
            set_txn_kwargs (dict | None): Additional transaction kwargs passed to
              :meth:`~ape.api.networks.EcosystemAPI.create_transaction` for the
              :meth:`AccountAPI.set_delegate` method, such as ``gas``, ``max_fee``, or
              ``max_priority_fee``. Overrides the values provided via ``txn_kwargs``. For a list of
              available transaction kwargs, see :class:`~ape.api.transactions.TransactionAPI`.
            reset_txn_kwargs (dict | None): Additional transaction kwargs passed to
              :meth:`~ape.api.networks.EcosystemAPI.create_transaction` for the
              :meth:`AccountAPI.remove_delegate` method, such as ``gas``, ``max_fee``, or
              ``max_priority_fee``. Overrides the values provided via ``txn_kwargs``. For a list of
              available transaction kwargs, see :class:`~ape.api.transactions.TransactionAPI`.
            **txn_kwargs: Additional transaction kwargs passed to
              :meth:`~ape.api.networks.EcosystemAPI.create_transaction`, such as ``gas``
              ``max_fee``, or ``max_priority_fee``. For a list of available transaction
              kwargs, see :class:`~ape.api.transactions.TransactionAPI`.

        Returns:
            `:class:~ape.contracts.ContractInstance`:
                The contract instance of this account with the interface of `contract`.
        """
        set_txn_kwargs = {**txn_kwargs, **(set_txn_kwargs or {})}
        existing_delegate = self.delegate

        self.set_delegate(new_delegate, **set_txn_kwargs)

        # NOTE: Do not cache this type as it is temporary
        from ape.contracts import ContractInstance

        # This is helpful for using it immediately to send things as self
        with self.account_manager.use_sender(self):
            if isinstance(new_delegate, ContractInstance):
                # NOTE: Do not cache this
                yield ContractInstance(self.address, contract_type=new_delegate.contract_type)

            else:
                yield self

        reset_txn_kwargs = {**txn_kwargs, **(reset_txn_kwargs or {})}
        if existing_delegate:
            self.set_delegate(existing_delegate, **reset_txn_kwargs)
        else:
            self.remove_delegate(**reset_txn_kwargs)


class AccountContainerAPI(BaseInterfaceModel):
    """
    An API class representing a collection of :class:`~ape.api.accounts.AccountAPI`
    instances.
    """

    name: str
    """
    The name of the account container.
    For example, the ``ape-ledger`` plugin
    uses ``"ledger"`` as its name.
    """

    account_type: type[AccountAPI]
    """
    The type of account in this container.
    See :class:`~ape.api.accounts.AccountAPI`.
    """

    @property
    @abstractmethod
    def aliases(self) -> Iterator[str]:
        """
        All available aliases.

        Returns:
            Iterator[str]
        """

    @property
    @abstractmethod
    def accounts(self) -> Iterator[AccountAPI]:
        """
        All accounts.

        Returns:
            Iterator[:class:`~ape.api.accounts.AccountAPI`]
        """

    @cached_property
    def data_folder(self) -> Path:
        """
        The path to the account data files.
        Defaults to ``$HOME/.ape/<plugin_name>`` unless overridden.
        """
        path = self.config_manager.DATA_FOLDER / self.name
        path.mkdir(parents=True, exist_ok=True)
        return path

    @abstractmethod
    def __len__(self) -> int:
        """
        Number of accounts.
        """

    def __getitem__(self, address: AddressType) -> AccountAPI:
        """
        Get an account by address.

        Args:
            address (:class:`~ape.types.address.AddressType`): The address to get. The type is an alias to
              `ChecksumAddress <https://eth-typing.readthedocs.io/en/latest/types.html#checksumaddress>`__.  # noqa: E501

        Raises:
            KeyError: When there is no local account with the given address.

        Returns:
            :class:`~ape.api.accounts.AccountAPI`
        """
        for account in self.accounts:
            if account.address == address:
                return account

        raise KeyError(f"No local account {address}.")

    def append(self, account: AccountAPI):
        """
        Add an account to the container.

        Raises:
            :class:`~ape.exceptions.AccountsError`: When the account is already in the container.

        Args:
            account (:class:`~ape.api.accounts.AccountAPI`): The account to add.
        """
        self._verify_account_type(account)

        if account.address in self:
            raise AccountsError(f"Account '{account.address}' already in container.")

        self._verify_unused_alias(account)

        self.__setitem__(account.address, account)

    def __setitem__(self, address: AddressType, account: AccountAPI):
        raise APINotImplementedError("Must define this method to use `container.append(acct)`.")

    def remove(self, account: AccountAPI):
        """
        Delete an account.

        Raises:
            :class:`~ape.exceptions.AccountsError`: When the account is not known to ``ape``.

        Args:
            account (:class:`~ape.accounts.AccountAPI`): The account to remove.
        """
        self._verify_account_type(account)

        if account.address not in self:
            raise AccountsError(f"Account '{account.address}' not known.")

        self.__delitem__(account.address)

    def __delitem__(self, address: AddressType):
        """
        Delete an account.

        Raises:
            NotImplementError: When not overridden within a plugin.

        Args:
            address (:class:`~ape.types.address.AddressType`): The address of the account to delete.
        """
        raise APINotImplementedError("Must define this method to use `container.remove(acct)`.")

    def __contains__(self, address: AddressType) -> bool:
        """
        Check if the address is an existing account in ``ape``.

        Raises:
            IndexError: When the given account address is not in this container.

        Args:
            address (:class:`~ape.types.address.AddressType`): An account address.

        Returns:
            bool: ``True`` if ``ape`` manages the account with the given address.
        """
        try:
            self.__getitem__(address)
            return True

        except (IndexError, KeyError, AttributeError):
            return False

    def _verify_account_type(self, account):
        if not isinstance(account, self.account_type):
            container_type_name = getattr(type(account), "__name__", "<CustomContainerType>")
            account_type_name = getattr(self.account_type, "__name__", "<UnknownAccount>")
            message = (
                f"Container '{container_type_name}' is an incorrect "
                f"type for container '{account_type_name}'."
            )

            raise AccountsError(message)

    def _verify_unused_alias(self, account):
        if account.alias and account.alias in self.aliases:
            raise AliasAlreadyInUseError(account.alias)


class TestAccountContainerAPI(AccountContainerAPI):
    """
    Test account containers for ``ape test`` (such containers that generate accounts using
    :class:`~ape.utils.GeneratedDevAccounts`) should implement this API instead of
    ``AccountContainerAPI`` directly. Then, they show up in the ``accounts`` test fixture.
    """

    @property
    def mnemonic(self) -> str:
        return self.config_manager.test.get("mnemonic", DEFAULT_TEST_MNEMONIC)

    @mnemonic.setter
    def mnemonic(self, value: str):
        self.config_manager.test.mnemonic = value

    @property
    def number_of_accounts(self) -> int:
        return self.config_manager.test.get("number_of_accounts", DEFAULT_NUMBER_OF_TEST_ACCOUNTS)

    @number_of_accounts.setter
    def number_of_accounts(self, value: int):
        self.config_manager.test.number_of_accounts = value

    @property
    def hd_path(self) -> str:
        return self.config_manager.test.get("hd_path", DEFAULT_TEST_HD_PATH)

    @hd_path.setter
    def hd_path(self, value: str):
        self.config_manager.test.hd_path = value

    @cached_property
    def data_folder(self) -> Path:
        """
        **NOTE**: Test account containers do not touch
        persistent data. By default and unless overridden,
        this property returns the path to ``/dev/null`` and
        it is not used for anything.
        """
        return Path("/dev/null" if os.name == "posix" else "NUL")

    @raises_not_implemented
    def get_test_account(self, index: int) -> "TestAccountAPI":  # type: ignore[empty-body]
        """
        Get the test account at the given index.

        Args:
            index (int): The index of the test account.

        Returns:
            :class:`~ape.api.accounts.TestAccountAPI`
        """

    @abstractmethod
    def generate_account(self, index: Optional[int] = None) -> "TestAccountAPI":
        """
        Generate a new test account.
        """

    def reset(self):
        """
        Reset the account container to an original state.
        """


class TestAccountAPI(AccountAPI):
    """
    Test accounts for ``ape test`` (such accounts that use
    :class:`~ape.utils.GeneratedDevAccounts`) should implement this API
    instead of ``AccountAPI`` directly. Then, they show up in the ``accounts`` test fixture.
    """


class ImpersonatedAccount(AccountAPI):
    """
    An account to use that does not require signing.
    """

    raw_address: AddressType
    """
    The field-address of the account.
    """

    @property
    def address(self) -> AddressType:
        return self.raw_address

    def sign_message(self, msg: Any, **signer_options) -> Optional[MessageSignature]:
        raise APINotImplementedError("This account cannot sign messages")

    def sign_transaction(self, txn: TransactionAPI, **signer_options) -> Optional[TransactionAPI]:
        # Returns input transaction unsigned (since it doesn't have access to the key)
        return txn

    def call(
        self,
        txn: TransactionAPI,
        send_everything: bool = False,
        private: bool = False,
        sign: bool = True,
        **kwargs,
    ) -> ReceiptAPI:
        txn = self.prepare_transaction(txn)
        txn.sender = txn.sender or self.raw_address

        return (
            self.provider.send_private_transaction(txn)
            if private
            else self.provider.send_transaction(txn)
        )



import json
from pathlib import Path as PathLibPath
from typing import Any

import click


class Path(click.Path):
    """
    This class exists to encourage the consistent usage
    of ``pathlib.Path`` for path_type.
    """

    def __init__(self, *args, **kwargs):
        if "path_type" not in kwargs:
            kwargs["path_type"] = PathLibPath

        super().__init__(*args, **kwargs)


class JSON(click.ParamType):
    """
    A type that accepts a raw-JSON str
    and loads it into a dictionary.
    """

    def convert(self, value, param, ctx):
        if not value:
            return {}

        elif isinstance(value, str):
            try:
                return json.loads(value)
            except ValueError as err:
                self.fail(f"Invalid JSON string: {err}", param, ctx)

        return value  # Good already.


class Noop(click.ParamType):
    """
    A param-type for ignoring param-types.
    Good to use when the multi-type handling
    happens already in a callback or in the command itself.
    """

    def convert(self, value: Any, param, ctx) -> Any:
        return value



import json
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING

import click
import yaml
from rich import print as echo_rich_text
from rich.tree import Tree

from ape.cli.choices import LazyChoice, OutputFormat
from ape.cli.options import ape_cli_context, network_option, output_format_option
from ape.exceptions import NetworkError
from ape.logging import LogLevel

if TYPE_CHECKING:
    from ape.api.providers import SubprocessProvider


def _filter_option(name: str, get_options: Callable[[], Sequence[str]]):
    return click.option(
        f"--{name}",
        f"{name}_filter",
        multiple=True,
        help=f"Filter the results by {name}",
        type=LazyChoice(get_options),
    )


@click.group(short_help="Manage networks")
def cli():
    """
    Command-line helper for managing networks.
    """


def _lazy_get(name: str) -> Sequence:
    # NOTE: Using fn generator to maintain laziness.
    def gen():
        from ape.utils.basemodel import ManagerAccessMixin as access

        yield from getattr(access.network_manager, f"{name}_names")

    from ape.types.basic import _LazySequence

    return _LazySequence(gen)


@cli.command(name="list", short_help="List registered networks")
@ape_cli_context()
@output_format_option()
@_filter_option("ecosystem", lambda: _lazy_get("ecosystem"))
@_filter_option("network", lambda: _lazy_get("network"))
@_filter_option("provider", lambda: _lazy_get("provider"))
@click.option("--running", is_flag=True, help="List running networks")
def _list(cli_ctx, output_format, ecosystem_filter, network_filter, provider_filter, running):
    """
    List all the registered ecosystems, networks, and providers.
    """
    if running:
        # TODO: Honor filter args.
        _print_running_networks(cli_ctx)
        return

    network_data = cli_ctx.network_manager.get_network_data(
        ecosystem_filter=ecosystem_filter,
        network_filter=network_filter,
        provider_filter=provider_filter,
    )

    if output_format == OutputFormat.TREE:
        default_suffix = "[dim default]  (default)"
        ecosystems = network_data["ecosystems"]
        ecosystems = sorted(ecosystems, key=lambda e: e["name"])

        def make_sub_tree(data: dict, create_tree: Callable) -> Tree:
            name = f"[bold green]{data['name']}"
            if data.get("isDefault"):
                name += default_suffix

            sub_tree = create_tree(name)
            return sub_tree

        for ecosystem in ecosystems:
            ecosystem_tree = make_sub_tree(ecosystem, Tree)
            _networks = {n["name"]: n for n in ecosystem["networks"]}
            _networks = {n: _networks[n] for n in sorted(_networks)}
            for network_name, network in _networks.items():
                providers = network["providers"]
                if providers:
                    network_tree = make_sub_tree(network, ecosystem_tree.add)
                    providers = sorted(providers, key=lambda p: p["name"])
                    for provider in providers:
                        make_sub_tree(provider, network_tree.add)

            if _networks and ecosystem_tree.children:
                echo_rich_text(ecosystem_tree)

    elif output_format == OutputFormat.YAML:
        if not isinstance(network_data, dict):
            raise TypeError(
                f"Unexpected network data type: {type(network_data)}. "
                f"Expecting dict. YAML dump will fail."
            )

        try:
            click.echo(yaml.safe_dump(network_data, sort_keys=True).strip())
        except ValueError as err:
            try:
                data_str = json.dumps(network_data)
            except Exception:
                data_str = str(network_data)

            raise NetworkError(
                f"Network data did not dump to YAML: {data_str}\nActual err: {err}"
            ) from err


def _print_running_networks(cli_ctx):
    from ape.utils.os import clean_path

    rows = [["PID", "NETWORK", "IPC", "HTTP", "WS"]]  # Store headers as a list
    for pid, node in cli_ctx.network_manager.running_nodes.nodes.items():
        rows.append(
            [
                pid,
                node.network_choice,
                str(clean_path(node.ipc_path) if node.ipc_path else None),
                node.http_uri,
                node.ws_uri,
            ]
        )

    if len(rows) == 1:
        # Only header row.
        click.echo("Local node(s) not running.")

    else:
        col_widths = [max(len(str(row[i])) for row in rows) for i in range(len(rows[0]))]
        for row in rows:
            formatted_row = "  ".join(str(row[i]).ljust(col_widths[i]) for i in range(len(row)))
            echo_rich_text(formatted_row)


@cli.command(short_help="Start a node process")
@ape_cli_context()
@network_option(default="ethereum:local:node")
@click.option("--block-time", default=None, type=int, help="Block time in seconds")
@click.option("--background", is_flag=True, help="Run in the background")
def run(cli_ctx, provider, block_time, background):
    """
    Start a subprocess node as if running independently
    and stream stdout and stderr.
    """
    from ape.api.providers import SubprocessProvider

    # Ignore extra loggers, such as web3 loggers.
    cli_ctx.logger._extra_loggers = {}
    if not isinstance(provider, SubprocessProvider):
        cli_ctx.abort(
            f"`ape networks run` requires a provider that manages a process, not '{provider.name}'."
        )
    elif provider.is_connected:
        cli_ctx.abort("Process already running.")

    # Set block time if provided
    if block_time is not None:
        provider.provider_settings.update({"block_time": block_time})

    # Start showing process logs.
    original_level = cli_ctx.logger.level
    original_format = cli_ctx.logger.fmt
    cli_ctx.logger.set_level(LogLevel.DEBUG)

    # Change format to exclude log level (since it is always just DEBUG)
    cli_ctx.logger.format(fmt="%(message)s")

    try:
        _run(cli_ctx, provider, background=background)
    finally:
        cli_ctx.logger.set_level(original_level)
        cli_ctx.logger.format(fmt=original_format)


def _run(cli_ctx, provider: "SubprocessProvider", background: bool = False):
    provider.background = background
    provider.connect()

    if process := provider.process:
        if background:
            # End this process, letting the node continue running.
            # This node can be killed later using the `ape networks kill` command.
            return

        try:
            process.wait()
        finally:
            try:
                provider.disconnect()
            except Exception:
                # Prevent not being able to CTRL-C.
                cli_ctx.abort("Terminated")

    else:
        provider.disconnect()
        cli_ctx.abort("Process already running.")


@cli.command(short_help="Stop node processes")
@ape_cli_context()
@click.argument("process_ids", nargs=-1, type=int)
@click.option("--all", "kill_all", is_flag=True, help="Kill all running processes")
@network_option(default=None)
def kill(cli_ctx, process_ids, kill_all):
    """
    Stop node processes
    """
    if kill_all:
        if process_ids:
            raise click.BadOptionUsage("--all", "Cannot use `--all` with PID arguments.")

        process_ids = cli_ctx.network_manager.running_nodes.process_ids

    if not process_ids:
        message = (
            "No running nodes found."
            if not cli_ctx.network_manager.running_nodes
            else "No processes given. Use `--all` to kill all processes."
        )
        echo_rich_text(message)
        return

    elif processes_killed := cli_ctx.network_manager.kill_node_process(*process_ids):
        # Killed 1 or more nodes.
        click.echo("Stopped the following node(s):")
        pids_stopped = set()
        for pid, data in processes_killed.items():
            echo_rich_text(f"\t{repr(data)}")
            pids_stopped.add(pid)

        if rest := [pid for pid in process_ids if pid not in pids_stopped]:
            click.echo(f"The remaining process IDs were no longer valid: {','.join(rest)}.")

    else:
        # Terminated the process outside of Ape.
        click.echo("No running nodes found, but cleaned up cache.")


@cli.command(short_help="Check if a provider is available")
@ape_cli_context()
@network_option()
def ping(cli_ctx, provider):
    if hasattr(provider, "allow_start"):
        # We don't want to allow starting processes; this is used for
        # checking if processes are alive (as well as checking live URIs).
        provider.allow_start = False

    provider.connect()
    status = "AVAILABLE" if provider.is_connected else "UNAVAILABLE"
    click.echo(f"'{provider.network_choice}' connection status: {status}")



def test_get_contract(benchmark, project_with_contracts):
    _ = project_with_contracts.Other  # Ensure compiled first.
    benchmark.pedantic(
        lambda *args, **kwargs: project_with_contracts.get_contract(*args, **kwargs),
        args=(("Other",),),
        rounds=5,
        warmup_rounds=1,  # It's always slower the first time, a little bit.
    )
    stats = benchmark.stats
    median = stats.get("median")

    # NOTE: At one point, this was average '0.0007'.
    # When I run locally, I tend to get 0.0001.
    # In CI, when very busy, it can get slower
    assert median < 0.00070



from eth_utils import to_hex

from ape.api import TraceAPI
from ape.utils import ManagerAccessMixin
from tests.conftest import geth_process_test


@geth_process_test
def test_return_value_list(geth_account, geth_contract):
    tx = geth_contract.getFilledArray.transact(sender=geth_account)
    assert tx.return_value == [1, 2, 3]


@geth_process_test
def test_return_value_nested_address_array(geth_account, geth_contract, zero_address):
    tx = geth_contract.getNestedAddressArray.transact(sender=geth_account)
    expected = [
        [geth_account.address, geth_account.address, geth_account.address],
        [zero_address, zero_address, zero_address],
    ]
    actual = tx.return_value
    assert actual == expected


@geth_process_test
def test_return_value_nested_struct_in_tuple(geth_account, geth_contract):
    tx = geth_contract.getNestedStructWithTuple1.transact(sender=geth_account)
    actual = tx.return_value
    assert actual[0].t.a == geth_account.address
    assert actual[0].foo == 1
    assert actual[1] == 1


@geth_process_test
def test_trace(geth_account, geth_contract):
    tx = geth_contract.getNestedStructWithTuple1.transact(sender=geth_account)
    assert isinstance(tx.trace, TraceAPI)


@geth_process_test
def test_track_gas(mocker, geth_account, geth_contract, gas_tracker):
    tx = geth_contract.getNestedStructWithTuple1.transact(sender=geth_account)
    mock_test_runner = mocker.MagicMock()
    mock_test_runner.gas_tracker = gas_tracker

    ManagerAccessMixin._test_runner = mock_test_runner

    try:
        tx.track_gas()
    finally:
        ManagerAccessMixin._test_runner = None

    report = gas_tracker.session_gas_report or {}
    contract_name = geth_contract.contract_type.name
    assert contract_name in report
    assert "getNestedStructWithTuple1" in report[contract_name]


@geth_process_test
def test_await_confirmations(geth_account, geth_contract):
    tx = geth_contract.setNumber(235921972943759, sender=geth_account)
    tx.await_confirmations()
    assert tx.confirmed


@geth_process_test
def test_await_confirmations_zero_confirmations(mocker, geth_account, geth_contract):
    """
    We still need to wait for the nonce to increase when required confirmations is 0.
    Otherwise, we sometimes ran into nonce-issues when transacting too fast with
    the same account.
    """
    tx = geth_contract.setNumber(545921972923759, sender=geth_account, required_confirmations=0)
    spy = mocker.spy(tx, "_await_sender_nonce_increment")
    tx.await_confirmations()
    assert tx.confirmed
    assert spy.call_count == 1


@geth_process_test
def test_transaction(geth_account, geth_contract):
    receipt = geth_contract.setNumber(1998, sender=geth_account)
    actual = receipt.transaction
    assert actual.sender == geth_account.address
    assert to_hex(actual.txn_hash) == receipt.txn_hash



# Inspired / borrowed from the `click-logging` python package.
import logging
import sys
import traceback
from collections.abc import Callable, Iterator, Sequence
from contextlib import contextmanager
from enum import IntEnum
from pathlib import Path
from typing import IO, TYPE_CHECKING, Any, Optional, Union
from urllib.parse import urlparse, urlunparse

import click

if TYPE_CHECKING:
    from rich.console import Console as RichConsole


class LogLevel(IntEnum):
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    SUCCESS = logging.INFO + 1
    INFO = logging.INFO
    DEBUG = logging.DEBUG


logging.addLevelName(LogLevel.SUCCESS.value, LogLevel.SUCCESS.name)
logging.SUCCESS = LogLevel.SUCCESS.value  # type: ignore
DEFAULT_LOG_LEVEL = LogLevel.INFO.name
DEFAULT_LOG_FORMAT = "%(levelname_semicolon_padded)s %(plugin)s %(message)s"
HIDDEN_MESSAGE = "[hidden]"


def success(self, message, *args, **kws):
    """This method gets injected into python's `logging` module
    to handle logging at this level."""

    if self.isEnabledFor(LogLevel.SUCCESS.value):
        # Yes, logger takes its '*args' as 'args'.
        self._log(LogLevel.SUCCESS.value, message, args, **kws)


logging.Logger.success = success  # type: ignore


CLICK_STYLE_KWARGS = {
    LogLevel.ERROR: dict(fg="bright_red"),
    LogLevel.WARNING: dict(fg="bright_yellow"),
    LogLevel.SUCCESS: dict(fg="bright_green"),
    LogLevel.INFO: dict(fg="blue"),
    LogLevel.DEBUG: dict(fg="blue"),
}
CLICK_ECHO_KWARGS = {
    LogLevel.ERROR: dict(err=True),
    LogLevel.WARNING: dict(err=True),
    LogLevel.SUCCESS: dict(),
    LogLevel.INFO: dict(),
    LogLevel.DEBUG: dict(),
}


# Borrowed from `click._compat`.
def _isatty(stream: IO) -> bool:
    """Returns ``True`` if the stream is part of a tty.
    Borrowed from ``click._compat``."""
    # noinspection PyBroadException
    try:
        return stream.isatty()
    except Exception:
        return False


class ApeColorFormatter(logging.Formatter):
    def __init__(self, fmt: Optional[str] = None):
        fmt = fmt or DEFAULT_LOG_FORMAT
        super().__init__(fmt=fmt)

    def format(self, record):
        record.levelname_semicolon_padded = f"{record.levelname}:".ljust(8)
        if _isatty(sys.stdout) and _isatty(sys.stderr):
            # Only color log messages when sys.stdout and sys.stderr are sent to the terminal.
            level = LogLevel(record.levelno)
            default_dict: dict[str, Any] = {}
            styles: dict[str, Any] = CLICK_STYLE_KWARGS.get(level, default_dict)
            record.levelname = click.style(record.levelname, **styles)
            record.levelname_semicolon_padded = click.style(
                record.levelname_semicolon_padded, **styles
            )

        path = Path(record.pathname)
        record.plugin = ""
        for part in path.parts:
            if part.startswith("ape-"):
                record.plugin = f" ({part})"
                break

        return super().format(record)


class ClickHandler(logging.Handler):
    def __init__(
        self, echo_kwargs: dict, handlers: Optional[Sequence[Callable[[str], str]]] = None
    ):
        super().__init__()
        self.echo_kwargs = echo_kwargs
        self.handlers = handlers or []

    def emit(self, record):
        try:
            msg = self.format(record)
            for handler in self.handlers:
                msg = handler(msg)

            level = record.levelname.lower()
            if self.echo_kwargs.get(level):
                click.echo(msg, **self.echo_kwargs[level])
            else:
                click.echo(msg)
        except Exception:
            self.handleError(record)


class ApeLogger:
    _mentioned_verbosity_option = False
    _extra_loggers: dict[str, logging.Logger] = {}
    DISABLE_LEVEL: int = 100_000

    def __init__(
        self,
        _logger: logging.Logger,
        fmt: str,
    ):
        self.error = _logger.error
        self.warning = _logger.warning
        self.success = getattr(_logger, "success", _logger.info)
        self.info = _logger.info
        self.debug = _logger.debug
        self._logger = _logger
        self._did_parse_sys_argv = False
        self._load_from_sys_argv()
        self.fmt = fmt

    @classmethod
    def create(cls, fmt: Optional[str] = None) -> "ApeLogger":
        fmt = fmt or DEFAULT_LOG_FORMAT
        _logger = get_logger("ape", fmt=fmt)
        return cls(_logger, fmt)

    def format(self, fmt: Optional[str] = None):
        self.fmt = fmt or DEFAULT_LOG_FORMAT
        fmt = fmt or DEFAULT_LOG_FORMAT
        _format_logger(self._logger, fmt)

    def _load_from_sys_argv(self, default: Optional[Union[str, int, LogLevel]] = None):
        """
        Load from sys.argv to beat race condition with `click`.
        """
        if self._did_parse_sys_argv:
            # Already parsed.
            return

        log_level = _get_level(level=default)
        level_names = [lvl.name for lvl in LogLevel]

        #  Minus 2 because if `-v` is the last arg, it is not our verbosity `-v`.
        num_args = len(sys.argv) - 2

        for arg_i in range(1, 1 + num_args):
            if sys.argv[arg_i] == "-v" or sys.argv[arg_i] == "--verbosity":
                try:
                    level = _get_level(sys.argv[arg_i + 1].upper())
                except Exception:
                    # Let it fail in a better spot, or is not our level.
                    continue

                if level in level_names:
                    self._sys_argv = level
                    log_level = level
                    break
                else:
                    # Not our level.
                    continue

        self.set_level(log_level)
        self._did_parse_sys_argv = True

    @property
    def level(self) -> int:
        return self._logger.level

    def set_level(self, level: Union[str, int, LogLevel]):
        """
        Change the global ape logger log-level.

        Args:
            level (str): The name of the level or the value of the log-level.
        """
        if level == self._logger.level:
            return
        elif isinstance(level, LogLevel):
            level = level.value
        elif isinstance(level, str) and level.lower().startswith("loglevel."):
            # Seen in some environments.
            level = level.split(".")[-1].strip()

        self._logger.setLevel(level)
        for _logger in self._extra_loggers.values():
            _logger.setLevel(level)

    @contextmanager
    def at_level(self, level: Union[str, int, LogLevel]) -> Iterator:
        """
        Change the log-level in a context.

        Args:
            level (Union[str, int, LogLevel]): The level to use.

        Returns:
            Iterator
        """
        initial_level = self.level
        self.set_level(level)
        yield
        self.set_level(initial_level)

    def disable(self):
        self.set_level(self.DISABLE_LEVEL)

    @contextmanager
    def disabled(self):
        with self.at_level(self.DISABLE_LEVEL):
            yield

    def log_error(self, err: Exception):
        """
        Avoids logging empty messages.
        """
        message = str(err)
        if message:
            self._logger.error(message)

    def warn_from_exception(self, err: Exception, message: str):
        """
        Warn the user with the given message,
        log the stack-trace of the error at the DEBUG level, and
        mention how to enable DEBUG logging (only once).
        """
        message = self._create_message_from_error(err, message)
        self._logger.warning(message)
        self.log_debug_stack_trace()

    def error_from_exception(self, err: Exception, message: str):
        """
        Log an error to the user with the given message,
        log the stack-trace of the error at the DEBUG level, and
        mention how to enable DEBUG logging (only once).
        """
        message = self._create_message_from_error(err, message)
        self._logger.error(message)
        self.log_debug_stack_trace()

    def _create_message_from_error(self, err: Exception, message: str):
        err_type_name = getattr(type(err), "__name__", "Exception")
        err_output = f"{err_type_name}: {err}"
        message = f"{message}\n\t{err_output}"
        if not self._mentioned_verbosity_option:
            message += "\n\t(Use `--verbosity DEBUG` to see full stack-trace)"
            self._mentioned_verbosity_option = True

        return message

    def log_debug_stack_trace(self):
        stack_trace = traceback.format_exc()
        self._logger.debug(stack_trace)

    def create_logger(
        self, new_name: str, handlers: Optional[Sequence[Callable[[str], str]]] = None
    ) -> logging.Logger:
        _logger = get_logger(new_name, fmt=self.fmt, handlers=handlers)
        _logger.setLevel(self.level)
        self._extra_loggers[new_name] = _logger
        return _logger


def _format_logger(
    _logger: logging.Logger, fmt: str, handlers: Optional[Sequence[Callable[[str], str]]] = None
):
    handler = ClickHandler(echo_kwargs=CLICK_ECHO_KWARGS, handlers=handlers)
    formatter = ApeColorFormatter(fmt=fmt)
    handler.setFormatter(formatter)

    # Remove existing handler(s)
    for existing_handler in _logger.handlers[:]:
        if isinstance(existing_handler, ClickHandler):
            _logger.removeHandler(existing_handler)

    _logger.addHandler(handler)


def get_logger(
    name: str,
    fmt: Optional[str] = None,
    handlers: Optional[Sequence[Callable[[str], str]]] = None,
) -> logging.Logger:
    """
    Get a logger with the given ``name`` and configure it for usage with Ape.

    Args:
        name (str): The name of the logger.
        fmt (Optional[str]): The format of the logger. Defaults to the Ape
          logger's default format: ``"%(levelname)s%(plugin)s: %(message)s"``.
        handlers (Optional[Sequence[Callable[[str], str]]]): Additional log message handlers.

    Returns:
        ``logging.Logger``
    """
    _logger = logging.getLogger(name)
    _format_logger(_logger, fmt=fmt or DEFAULT_LOG_FORMAT, handlers=handlers)
    return _logger


def _get_level(level: Optional[Union[str, int, LogLevel]] = None) -> str:
    if level is None:
        return DEFAULT_LOG_LEVEL
    elif isinstance(level, LogLevel):
        return level.name
    elif isinstance(level, int) or (isinstance(level, str) and level.isnumeric()):
        return LogLevel(int(level)).name
    elif isinstance(level, str) and level.lower().startswith("loglevel."):
        # Handle 'LogLevel.' prefix.
        return level.split(".")[-1].strip()

    return level


def sanitize_url(url: str) -> str:
    """Removes sensitive information from given URL"""

    parsed = urlparse(url)
    new_netloc = parsed.hostname or ""
    if parsed.port:
        new_netloc += f":{parsed.port}"

    new_url = urlunparse(parsed._replace(netloc=new_netloc, path=""))
    return f"{new_url}/{HIDDEN_MESSAGE}" if parsed.path else new_url


logger = ApeLogger.create()


class _RichConsoleFactory:
    rich_console_map: dict[str, "RichConsole"] = {}

    def get_console(self, file: Optional[IO[str]] = None, **kwargs) -> "RichConsole":
        # Configure custom file console
        file_id = str(file)
        if file_id not in self.rich_console_map:
            # perf: delay importing from rich, as it is slow.
            from rich.console import Console as RichConsole

            self.rich_console_map[file_id] = RichConsole(file=file, width=100, **kwargs)

        return self.rich_console_map[file_id]


_factory = _RichConsoleFactory()


def get_rich_console(file: Optional[IO[str]] = None, **kwargs) -> "RichConsole":
    """
    Get an Ape-configured rich console.

    Args:
        file (Optional[IO[str]]): The file to output to. Will default
          to using stdout.

    Returns:
        ``rich.Console``.
    """
    return _factory.get_console(file, **kwargs)


def silenced(func: Callable):
    """
    A decorator for ensuring a function does not output any logs.

    Args:
        func (Callable): The function to call silently.
    """

    def wrapper(*args, **kwargs):
        with logger.disabled():
            return func(*args, **kwargs)

    return wrapper


__all__ = ["DEFAULT_LOG_LEVEL", "ApeLogger", "LogLevel", "get_rich_console", "logger", "silenced"]



import time
from queue import Queue
from typing import TYPE_CHECKING, Optional

import pytest
from eth_pydantic_types import HexBytes, HexBytes20
from eth_utils import to_hex
from ethpm_types import ContractType

from ape.contracts.base import ContractEventWrapper
from ape.exceptions import ProviderError
from ape.types.events import ContractLog, MockContractLog
from ape.types.units import CurrencyValueComparable

if TYPE_CHECKING:
    from ape.api.transactions import ReceiptAPI


@pytest.fixture
def assert_log_values(owner):
    def _assert_log_values(log: ContractLog, number: int, previous_number: Optional[int] = None):
        assert isinstance(log.b, bytes)
        expected_previous_number = number - 1 if previous_number is None else previous_number
        assert log.prevNum == expected_previous_number, "Event param 'prevNum' has unexpected value"
        assert log.newNum == number, "Event param 'newNum' has unexpected value"
        assert log.newNum == f"{number} wei", "string comparison with number not working"
        assert log.dynData == "Dynamic"
        assert log.dynIndexed == HexBytes(
            "0x9f3d45ac20ccf04b45028b8080bb191eab93e29f7898ed43acf480dd80bba94d"
        )

    return _assert_log_values


def test_from_receipt(owner, contract_instance, assert_log_values):
    event_type = contract_instance.NumberChange

    # Invoke a transaction 3 times that generates 3 logs.
    receipt_0 = contract_instance.setNumber(1, sender=owner)
    receipt_1 = contract_instance.setNumber(2, sender=owner)
    receipt_2 = contract_instance.setNumber(3, sender=owner)

    def assert_receipt_logs(receipt: "ReceiptAPI", num: int):
        logs = event_type.from_receipt(receipt)
        assert len(logs) == 1
        assert_log_values(logs[0], num)

        # Also verify can we decode logs the other way
        logs = receipt.decode_logs(event_type)
        assert len(logs) == 1
        assert_log_values(logs[0], num)

    assert_receipt_logs(receipt_0, 1)
    assert_receipt_logs(receipt_1, 2)
    assert_receipt_logs(receipt_2, 3)


def test_from_receipt_different_contract(
    owner, vyper_contract_instance, solidity_contract_instance, assert_log_values
):
    # NOTE: Ths event type is similar but defined on a different contract.
    event_type = solidity_contract_instance.NumberChange
    receipt = vyper_contract_instance.setNumber(1, sender=owner)
    logs = event_type.from_receipt(receipt)
    assert len(logs) == 0


def test_from_event_type(contract_instance, owner, assert_log_values):
    event_type = contract_instance.NumberChange
    start_num = 6
    size = 20
    num_range = range(start_num, start_num + size)

    # Generate 20 logs
    for i in num_range:
        contract_instance.setNumber(i, sender=owner)

    # Collect 20 logs
    logs = [log for log in event_type]

    assert len(logs) == size, "Unexpected number of logs"
    for num, log in zip(num_range, logs):
        if num == start_num:
            assert_log_values(log, num, previous_number=0)
        else:
            assert_log_values(log, num)


def test_index_access(contract_instance, owner, assert_log_values):
    event_type = contract_instance.NumberChange

    contract_instance.setNumber(1, sender=owner)
    contract_instance.setNumber(2, sender=owner)
    contract_instance.setNumber(3, sender=owner)

    assert event_type[0] == contract_instance.NumberChange(None, 0, None, 1)
    assert event_type[1] == contract_instance.NumberChange(None, 1, newNum=2)
    assert event_type[2] == contract_instance.NumberChange(newNum=3, prevNum=2)

    # Verify negative index access
    assert event_type[-3] == contract_instance.NumberChange(None, 0, None, 1)
    assert event_type[-2] == contract_instance.NumberChange(None, 1, newNum=2)
    assert event_type[-1] == contract_instance.NumberChange(newNum=3, prevNum=2)


def test_splicing(contract_instance, owner, assert_log_values):
    event_type = contract_instance.NumberChange

    contract_instance.setNumber(1, sender=owner)
    contract_instance.setNumber(2, sender=owner)
    contract_instance.setNumber(3, sender=owner)

    logs = event_type[:2]
    assert len(logs) == 2
    assert_log_values(logs[0], 1)
    assert_log_values(logs[1], 2)

    logs = event_type[2:]
    assert len(logs) == 1
    assert_log_values(logs[0], 3)

    log = event_type[1]
    assert_log_values(log, 2)


def test_range(chain, contract_instance, owner, assert_log_values):
    contract_instance.setNumber(1, sender=owner)
    start = chain.blocks.height
    logs = [
        log
        for log in contract_instance.NumberChange.range(
            start, start + 100, search_topics={"newNum": 1}
        )
    ]
    assert len(logs) == 1, "Unexpected number of logs"
    assert_log_values(logs[0], 1)


def test_range_by_address(
    mocker,
    chain,
    eth_tester_provider,
    accounts,
    contract_instance,
    owner,
    assert_log_values,
):
    get_logs_spy = mocker.spy(eth_tester_provider.tester.ethereum_tester, "get_logs")
    contract_instance.setAddress(accounts[1], sender=owner)
    height = chain.blocks.height
    logs = [
        log
        for log in contract_instance.AddressChange.range(
            height, height + 1, search_topics={"newAddress": accounts[1]}
        )
    ]

    # NOTE: This spy assertion tests against a bug where address queries were not
    # 0x-prefixed. However, this was still valid in EthTester and thus was not causing
    # test failures.
    height_arg = chain.blocks.height
    actual = get_logs_spy.call_args[-1]
    expected = {
        "address": [contract_instance.address],
        "from_block": height_arg,
        "to_block": height_arg,
        "topics": [
            "0x7ff7bacc6cd661809ed1ddce28d4ad2c5b37779b61b9e3235f8262be529101a9",
            "0x00000000000000000000000070997970c51812dc3a010c7d01b50e0d17dc79c8",
        ],
    }
    assert actual == expected
    assert logs == [contract_instance.AddressChange(newAddress=accounts[1])]


def test_range_multiple_addresses(
    chain, contract_instance, contract_container, owner, assert_log_values
):
    another_instance = contract_container.deploy(0, sender=owner)
    start_block = chain.blocks.height
    contract_instance.setNumber(1, sender=owner)
    another_instance.setNumber(1, sender=owner)
    logs = list(
        contract_instance.NumberChange.range(
            start_block,
            start_block + 100,
            search_topics={"newNum": 1},
            extra_addresses=[another_instance.address],
        )
    )
    assert len(logs) == 2, f"Unexpected number of logs: {len(logs)}"
    assert logs[0] == contract_instance.NumberChange(newNum=1, prevNum=0)
    assert logs[1] == another_instance.NumberChange(newNum=1, prevNum=0)


def test_range_start_and_stop(contract_instance, owner, chain):
    # Create 1 event
    contract_instance.setNumber(1, sender=owner)

    # Grab start block after first event
    start_block = chain.blocks.height

    contract_instance.setNumber(2, sender=owner)
    contract_instance.setNumber(3, sender=owner)

    stop = start_block + 30  # Stop can be bigger than height, it doesn't not matter
    logs = [log for log in contract_instance.NumberChange.range(start_block, stop=stop)]
    assert len(logs) == 3, "Unexpected number of logs"


def test_range_only_stop(contract_instance, owner, chain):
    # Create 3 events
    start = chain.blocks.height
    contract_instance.setNumber(1, sender=owner)
    contract_instance.setNumber(2, sender=owner)
    contract_instance.setNumber(3, sender=owner)

    stop = start + 100  # Stop can be bigger than height, it doesn't not matter
    logs = [log for log in contract_instance.NumberChange.range(stop)]
    assert len(logs) >= 3, f"Unexpected number of logs: {len(logs)}"


def test_range_negative_start(contract_instance, owner):
    # Create 2 events
    contract_instance.setNumber(1, sender=owner)
    contract_instance.setNumber(2, sender=owner)
    logs = [log for log in contract_instance.NumberChange.range(-2, 0)]
    assert len(logs) == 2


def test_range_negative_start_and_stop(contract_instance, owner):
    # Create 3 events
    contract_instance.setNumber(1, sender=owner)
    contract_instance.setNumber(2, sender=owner)
    contract_instance.setNumber(3, sender=owner)

    query_result = [log for log in contract_instance.NumberChange.range(-1, 0)]
    assert len(query_result) == 1, "Should only be 1"
    assert query_result[0].newNum == 3  # Was the last parameter.
    query_result = [log for log in contract_instance.NumberChange.range(-2, -1)]
    assert len(query_result) == 1, "Should only be 1"
    assert query_result[0].newNum == 2  # Was the penultimate parameter.
    query_result = [log for log in contract_instance.NumberChange.range(-3, -2)]
    assert len(query_result) == 1, "Should only be 1"
    assert query_result[0].newNum == 1  # Was the penultimate parameter.
    logs = [log for log in contract_instance.NumberChange.range(-3, -1)]
    assert len(logs) == 2
    assert [x.newNum for x in logs] == [1, 2]
    logs = [log for log in contract_instance.NumberChange.range(-3, 0)]
    assert len(logs) == 3
    assert [x.newNum for x in logs] == [1, 2, 3]


def test_range_negative_stop_only(contract_instance, owner):
    # Create 2 events
    contract_instance.setNumber(1, sender=owner)
    contract_instance.setNumber(2, sender=owner)

    # Get _all_ logs.
    logs = [log for log in contract_instance.NumberChange.range(0)]
    assert len(logs) == 2
    assert [x.newNum for x in logs] == [1, 2]

    # Basically means go from 0 to the second to last
    logs = [log for log in contract_instance.NumberChange.range(-1)]
    assert len(logs) == 1
    assert logs[0].newNum == 1


def test_poll_logs_stop_block_not_in_future(
    chain_that_mined_5, vyper_contract_instance, eth_tester_provider
):
    bad_stop_block = chain_that_mined_5.blocks.height

    with pytest.raises(ValueError, match="'stop' argument must be in the future."):
        _ = [x for x in vyper_contract_instance.NumberChange.poll_logs(stop_block=bad_stop_block)]


def test_poll_logs(chain, vyper_contract_instance, eth_tester_provider, owner, PollDaemon):
    size = 3
    logs: Queue = Queue(maxsize=size)
    start_block = vyper_contract_instance.creation_metadata.block or 0
    poller = vyper_contract_instance.NumberChange.poll_logs(start_block=start_block)
    start_block = chain.blocks.height

    with PollDaemon("logs", poller, logs.put, logs.full):
        # Sleep first to ensure listening before emitting logs.
        time.sleep(1)

        vyper_contract_instance.setNumber(1, sender=owner)  # block s+1
        vyper_contract_instance.setNumber(33, sender=owner)  # block s+2
        vyper_contract_instance.setNumber(7, sender=owner)  # block s+3

    actual = [logs.get() for _ in range(size)]
    assert all(a.newNum == e for a, e in zip(actual, (1, 33, 7)))
    assert actual[0].block_number == actual[0].block.number == start_block + 1
    assert actual[1].block_number == actual[1].block.number == actual[0].block_number + 1
    assert actual[2].block_number == actual[2].block.number == actual[1].block_number + 1


def test_poll_logs_with_topics(vyper_contract_instance, eth_tester_provider, owner, PollDaemon):
    size = 3
    logs: Queue = Queue(maxsize=size)
    start_block = vyper_contract_instance.creation_metadata.block or 0
    poller = vyper_contract_instance.NumberChange.poll_logs(start_block=start_block, newNum=33)

    with PollDaemon("logs-with-topics", poller, logs.put, logs.full):
        # Sleep first to ensure listening before emitting logs.
        time.sleep(1)

        # NOTE: Doing the same number every time because poller is not the greatest.
        vyper_contract_instance.setNumber(33, sender=owner)  # block s+1
        vyper_contract_instance.setNumber(33, sender=owner)  # block s+2
        vyper_contract_instance.setNumber(33, sender=owner)  # block s+3f

    actual = [logs.get() for _ in range(size)]
    assert all(a.newNum == e for a, e in zip(actual, (33,)))


def test_poll_logs_timeout(vyper_contract_instance, eth_tester_provider, owner, PollDaemon):
    new_block_timeout = 1
    poller = vyper_contract_instance.NumberChange.poll_logs(new_block_timeout=new_block_timeout)

    with pytest.raises(ProviderError) as err:
        with PollDaemon("logs-timeout", poller, lambda x: None, lambda: False):
            time.sleep(1.5)

    assert "Timed out waiting for next block" in str(err.value)


def test_contract_two_events_with_same_name(
    owner, chain, networks_connected_to_tester, shared_contracts_folder
):
    interface_path = shared_contracts_folder / "Interface.json"
    impl_path = shared_contracts_folder / "InterfaceImplementation.json"
    interface_text = interface_path.read_text(encoding="utf8")
    impl_text = impl_path.read_text(encoding="utf8")
    interface_contract_type = ContractType.model_validate_json(interface_text)
    impl_contract_type = ContractType.model_validate_json(impl_text)
    event_name = "FooEvent"

    # Ensure test is setup correctly in case scenario-data changed on accident
    assert len([e for e in impl_contract_type.events if e.name == event_name]) == 2
    assert len([e for e in interface_contract_type.events if e.name == event_name]) == 1

    impl_container = chain.contracts.get_container(impl_contract_type)
    impl_instance = owner.deploy(impl_container)

    # Show some features still work when referencing by __getattr__.
    wrapper = impl_instance.FooEvent
    assert isinstance(wrapper, ContractEventWrapper)
    mock_log = wrapper(bar=16)
    assert isinstance(mock_log, MockContractLog)

    expected_sig_from_impl = "FooEvent(uint256 bar, uint256 baz)"
    expected_sig_from_interface = "FooEvent(uint256 bar)"
    event_from_impl_contract = impl_instance.get_event_by_signature(expected_sig_from_impl)
    assert event_from_impl_contract.abi.signature == expected_sig_from_impl
    event_from_interface = impl_instance.get_event_by_signature(expected_sig_from_interface)
    assert event_from_interface.abi.signature == expected_sig_from_interface


def test_contract_decode_logs_no_abi(owner, contract_instance):
    receipt = contract_instance.setNumber(1, sender=owner)
    events = list(receipt.decode_logs())  # no abi
    assert len(events) == 1
    assert events == [contract_instance.NumberChange(newNum=1)]


def test_contract_decode_logs_falsy_check(owner, vyper_contract_instance):
    """
    Verifies a bug fix where false-y values differing were ignored.
    """

    tx = vyper_contract_instance.setNumber(1, sender=owner)
    with pytest.raises(AssertionError):
        assert tx.events == [vyper_contract_instance.NumberChange(newNum=0)]


def test_contract_log_container(owner, contract_instance):
    receipt = contract_instance.setNumber(1, sender=owner)
    events = receipt.events.filter(contract_instance.NumberChange, newNum=1)
    assert len(events) == 1


def test_filter_events_with_same_abi(
    owner, contract_with_call_depth, middle_contract, leaf_contract
):
    """
    Test shows that if we have a contract method emit multiple events with the
    same ABI, that each event only appears on the respective contract's
    filtering. This test verifies we filter by contract address as well as ABI.
    """

    tx = contract_with_call_depth.emitLogWithSameInterfaceFromMultipleContracts(sender=owner)

    assert contract_with_call_depth.OneOfMany(addr=owner.address) in tx.events
    assert middle_contract.OneOfMany(addr=contract_with_call_depth.address) in tx.events
    assert leaf_contract.OneOfMany(addr=contract_with_call_depth.address) in tx.events

    # Ensure each contract's event appears only once
    result_a = tx.events.filter(contract_with_call_depth.OneOfMany)
    assert result_a == [contract_with_call_depth.OneOfMany(addr=owner.address)]

    result_b = tx.events.filter(middle_contract.OneOfMany)
    assert result_b == [middle_contract.OneOfMany(addr=contract_with_call_depth.address)]

    result_c = tx.events.filter(leaf_contract.OneOfMany)
    assert result_c == [leaf_contract.OneOfMany(addr=contract_with_call_depth.address)]


def test_structs_in_events(contract_instance, owner, mystruct_c):
    tx = contract_instance.logStruct(sender=owner)
    expected_bytes = HexBytes(0x1234567890ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF)
    expected = contract_instance.EventWithStruct(
        a_struct={"a": owner, "b": expected_bytes, "c": mystruct_c}
    )
    assert tx.events == [expected]


def test_address_arrays_in_events(contract_instance, owner):
    tx = contract_instance.logAddressArray(sender=owner)
    expected = contract_instance.EventWithAddressArray(
        some_id=1001, some_address=owner, participants=[owner], agents=[owner]
    )
    assert tx.events == [expected]


def test_uint_arrays_in_events(contract_instance, owner):
    tx = contract_instance.logUintArray(sender=owner)
    expected = contract_instance.EventWithUintArray(agents=[1])
    assert tx.events == [expected]


def test_info(solidity_contract_instance):
    event_type = solidity_contract_instance.NumberChange
    actual = event_type.info
    header = (
        "NumberChange(bytes32 b, uint256 prevNum, string dynData, "
        "uint256 indexed newNum, string indexed dynIndexed)"
    )
    spec = (
        "@details Emitted when number is changed. `newNum` is the new number "
        "from the call. Expected every time number changes."
    )
    expected = f"""
{header}
  {spec}
""".strip()
    assert actual == expected


def test_model_dump(solidity_contract_container, owner):
    # NOTE: deploying a new contract with a new number to lessen x-dist conflicts.
    contract = owner.deploy(solidity_contract_container, 29620000000003)

    # First, get an event (a normal way).
    number = int(10e18)
    tx = contract.setNumber(number, sender=owner)
    event = tx.events[0]

    # Next, invoke `.model_dump()` to get the serialized version.
    log = event.model_dump()
    actual = log["event_arguments"]
    assert actual["newNum"] == number

    # This next assertion is important because of this Pydantic bug:
    # https://github.com/pydantic/pydantic/issues/10152
    assert not isinstance(actual["newNum"], CurrencyValueComparable)


@pytest.mark.parametrize("mode", ("python", "json"))
def test_model_dump_hexbytes(mode):
    # NOTE: There was an issue when using HexBytes for Any.
    event_arguments = {"key": 123, "validators": [HexBytes(123)]}
    txn_hash = HexBytes20.__eth_pydantic_validate__(347374237412374174)
    event = ContractLog(
        block_number=123,
        block_hash="block-hash",
        event_arguments=event_arguments,
        event_name="MyEvent",
        log_index=0,
        transaction_hash=txn_hash,
    )
    actual = event.model_dump(mode=mode)
    expected_hash = txn_hash if mode == "python" else to_hex(txn_hash)
    assert actual["transaction_hash"] == expected_hash


def test_model_dump_json():
    # NOTE: There was an issue when using HexBytes for Any.
    event_arguments = {"key": 123, "validators": [HexBytes(123)]}
    event = ContractLog(
        block_number=123,
        block_hash="block-hash",
        event_arguments=event_arguments,
        event_name="MyEvent",
        log_index=0,
        transaction_hash=HexBytes20.__eth_pydantic_validate__(347374237412374174),
    )
    actual = event.model_dump_json()
    assert actual == (
        '{"block_hash":"block-hash","block_number":123,'
        '"contract_address":"0x0000000000000000000000000000000000000000",'
        '"event_arguments":{"key":123,"validators":["0x7b"]},"event_name":'
        '"MyEvent","log_index":0,'
        '"transaction_hash":"0x00000000000000000000000004d21f074916369e"}'
    )


def test_transaction_hash():
    event_arguments = {"key": 123, "validators": [HexBytes(123)]}
    txn_hash = "a3430927834bd23"
    event = ContractLog(
        block_number=123,
        block_hash="block-hash",
        event_arguments=event_arguments,
        event_name="MyEvent",
        log_index=0,
        transaction_hash=txn_hash,
    )
    assert event.transaction_hash == f"0x{txn_hash}"



from pydantic_settings import SettingsConfigDict

from ape.api.config import PluginConfig


class CacheConfig(PluginConfig):
    size: int = 1024**3  # 1gb
    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_CACHE_")
